/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Sortov (sortov@ispras.ru)
// SYNOPSIS: Support of configuration files
///////////////////////////////////////////////////////////////////////////////

#include <fstream>
#include <sstream>

#include <utils/assertion.h>
#include <ts/config.hpp>

#define BUFSIZE 255

namespace cpptesk {
namespace ts {

Config::Config() {}

void Config::add_option(const std::string &name, const std::string &type, void *ptr)
{
    Option option(type, ptr);
    options[name] = option;
}

void Config::add_option(bool  &option, const std::string &name, bool  default_value)
{
    add_option(name, "bool", &option);
    option = default_value;
}

void Config::add_option(int   &option, const std::string &name, int   default_value)
{
    add_option(name, "int", &option);
    option = default_value;
}

void Config::add_option(uint  &option, const std::string &name, uint  default_value)
{
    add_option(name, "uint", &option);
    option = default_value;
}

void Config::add_option(float &option, const std::string &name, float default_value)
{
    add_option(name, "float", &option);
    option = default_value;
}


void Config::parse(const std::string &filename)
{
    std::ifstream config_file(filename.c_str(), std::ifstream::in);
    std::string line, line_wo_comments;

    std::string config_parse_msg("Config::parse: ");
    std::string file_error_msg("Can't open config file ");
    std::string opt_error_msg ("Unrecognized option ");
    std::string beg_bool_error_msg("Incorrect value (");
    std::string end_bool_error_msg(") of bool option ");

    unsigned linenum = 0;

    file_error_msg += filename;
    assertion(config_file != 0, file_error_msg.c_str());

    while (!config_file.eof())
    {
        char key[BUFSIZE];
        char val[BUFSIZE];
        char include[BUFSIZE];
        int res;

        std::getline (config_file, line);
        ++linenum;

        // ignore comments
        unsigned found = line.find("//");
        if (found != std::string::npos)
        {
            line_wo_comments = line.substr(0, found);
        }
        else
        {
            line_wo_comments = line;
        }

        // parse new file in #include
        if(1 == (res = sscanf(line_wo_comments.c_str(), "#include \"%[^\"]s\"", include)))
        {
            Config::parse(include);
        }

        // parse lines with key = val, ignore others
        if(2 != (res = sscanf(line_wo_comments.c_str(), "%[^=]=%[^\n]%*1c", key, val))) continue;

        // trim spaces
        char *t = key, *k = key;
        while (isspace(*t)) ++t;
        while (*t) *k++ = *t++;
        *k = '\0';
        while(isspace(*--k)) *k = '\0';

        std::map<std::string, Option>::iterator i = options.find(key);

        if( i != options.end() )
        {
            if(i->second.type == "bool")
            {
                int value = atoi(val);

                std::stringstream error_msg;
                error_msg << config_parse_msg + filename << ":" << linenum << ": "
                          << beg_bool_error_msg << value << end_bool_error_msg << i->first;
                assertion(value == 0 || value == 1, error_msg.str().c_str());
                *(bool *)i->second.ptr = value;
            }
            else if(i->second.type == "int")
            {
                int value = atoi(val);
                *(int *)i->second.ptr = value;
            }
            else if(i->second.type == "uint")
            {
                uint value = atoi(val);
                *(uint *)i->second.ptr = value;
            }
            else if(i->second.type == "float")
            {
                float value = atof(val);
                *(float *)i->second.ptr = value;
            }
        }
        else
        {
            std::stringstream error_msg;
            error_msg << config_parse_msg + filename << ":" << linenum << ": " << opt_error_msg + key;

            assertion(false, error_msg.str().c_str());
        }
    }

    return;
}

Config::~Config() {}

}}
