#pragma once

#include <c_tracer/c_tracer.h>
#include <ts/coverage.h>
#include <ts/defs.hpp>
#include <utils/utils.hpp>
#include <stdexcept>
#include <map>
#include <algorithm>
#include <vector>
#include <iostream>
#include <string>
#include <sstream>
#include <cstring>
#include <cstdlib>
#ifndef _MSC_VER
// For snprintf
#include <cstdio>
#endif

namespace cpptesk
{
namespace ts
{
namespace coverage
{

template<typename Id>
struct CoverageElement
{
    Id id;
    // int order;
    std::string label;
    CoverageElement(const Id& _id, const char* _label) : id(_id), label(_label)
    {
//        std::cout << "DEBUG: CoverageElement(const Id& _id, const char* _label): this == " << this << ", id == " << id << std::endl;
    }
    CoverageElement(const Id& _id, const std::string& _label) : id(_id), label(_label)
    {
//        std::cout << "DEBUG: CoverageElement(const Id& _id, const string& _label): this == " << this << ", id == " << id << std::endl;
    }


    CoverageElement(const CoverageElement& other) : id(other.id), label(other.label)
    {
//        std::cout << "DEBUG: CoverageElement(const CoverageElement&): this == " << this << ", other == " << &other << ", other.id == " << other.id << std::endl;
    }

    ~CoverageElement()
    {
//        std::cout << "DEBUG: ~CoverageElement(): this == " << this << ", id == " << id << std::endl;
    }
    CoverageElement& operator= (const CoverageElement& other)
    {
//        std::cout << "DEBUG: CoverageElement::operator= : this == " << this << ", other == " << &other << ", other.id == " << id << std::endl;
        {id = other.id; label = other.label; return *this; }
    }

    void swap(CoverageElement& other)
    {
//        std::cout << "DEBUG: CoverageElement::swap: this == " << this << ", this.id == " << id << ", other == " << &other << ", other.id == " << other.id << std::endl;
        std::swap(id, other.id);
        std::swap(label, other.label);
    }
};

template<typename Element, typename IdIterator, typename LabelIterator>
void populateElementRange(std::vector<Element>& dst, IdIterator id_begin, IdIterator id_end,
                          LabelIterator label_begin, LabelIterator label_end )
{
    while(id_begin != id_end && label_begin != label_end)
    {
        dst.push_back(Element(*id_begin, *label_begin));
        ++id_begin;
        ++label_begin;
    }
}

template<typename _Id1, typename _Id2>
struct ComposedId
{
    typedef _Id1 Id1;
    typedef _Id2 Id2;

    Id1 id1;
    Id2 id2;
    ComposedId() : id1(), id2() {}
    ComposedId(const Id1& r1, const Id2& r2) : id1(r1), id2(r2) {}
    template<typename U, typename V>
    ComposedId(const ComposedId<U,V>& r) : id1(r.id1), id2(r.id2) {}
    template<typename U, typename V>
    ComposedId& operator= (const ComposedId<U,V>& other)
    {
        id1 = other.id1;
        id2 = other.id2;
        return *this;
    }

    template<typename U, typename V>
    bool operator== (const ComposedId<U,V>& other) const
    {
        return id1 == other.id1 && id2 == other.id2;
    }
    template<typename U, typename V>
    bool operator!= (const ComposedId<U,V>& other) const
    {
        return id1 != other.id1 || id2 != other.id2;
    }
    template<typename U, typename V>
    bool operator < (const ComposedId<U,V>& other) const
    {
        return id1 < other.id1 || (id1==other.id1 && id2 < other.id2);
    }
    template<typename U, typename V>
    bool operator > (const ComposedId<U,V>& other) const
    {
        return id1 > other.id1 || (id1==other.id1 && id2 > other.id2);
    }
};

template<typename T1, typename T2>
std::ostream& operator << (std::ostream& out, const ComposedId<T1, T2>& id)
{
    return out << id.id1 << '*' << id.id2  ;
}

template<typename _Elements, typename _Name>
struct Coverage : public _Elements
{
    typedef _Elements Elements;
    typedef typename _Elements::Id Id;
    typedef typename _Elements::Element Element;
    typedef typename _Elements::ElementIterator ElementIterator;
    typedef Coverage _ThisType;

    static const char* name()
    {
        return _Name::name();
    }
    static size_t card()
    {
        return _Elements::get().size();
    }
    static ElementIterator begin()
    {
        return _Elements::get().begin();
    }
    static ElementIterator end()
    {
        return _Elements::get().end();
    }

    Id id;
    Coverage() : id() {}
    Coverage(const Coverage& other) : id(other.id) {}
    Coverage(const Id& _id) : id(_id) {}

    template<typename OtherCoverage>
    Coverage(const OtherCoverage& other) : id(other.id) {}

    ~Coverage() {}

    Coverage& operator= (const Id& _id)
    {
        id = _id;
        return *this;
    }
    Coverage& operator= (const Coverage& other)
    {
        id = other.id;
        return *this;
    }
    template<typename OtherCoverage>
    Coverage& operator= (const OtherCoverage& other)
    {
        id = other.id;
        return *this;
    }

    operator Id ()
    {
        return id;
    }

    bool operator == (const Coverage& o) const
    {
        return id == o.id;
    }
    bool operator == (const Id& o) const
    {
        return id == o;
    }
    bool operator != (const Coverage& o) const
    {
        return id != o.id;
    }
    bool operator != (const Id& o) const
    {
        return id != o;
    }
    bool operator < (const Coverage& o) const
    {
        return id < o.id;
    }
    bool operator < (const Id& o) const
    {
        return id < o;
    }
    bool operator > (const Coverage& o) const
    {
        return id > o.id;
    }
    bool operator > (const Id& o) const
    {
        return id > o;
    }

    template<typename T>
    inline ComposedId<Id, typename T::Id> operator*(const T& o)
    {
        return ComposedId<Id, typename T::Id>(id, o.id);
    }

    struct Matcher
    {
        typename _Elements::Matcher m;
        inline bool operator() (const Coverage& c) const
        {
            return m(c.id);
        }
        inline bool operator() (const Id& id) const
        {
            return m(id);
        }
        Matcher& operator= (const Id& id)
        {
            m = id;
            return *this;
        }
        Matcher& operator= (const Coverage& c)
        {
            m = c.id;
            return *this;
        }
    };
    static inline Matcher ANY()
    {
        return Matcher();
    }
    inline operator Matcher()
    {
        Matcher result = id;
        return result;
    }
};

template <typename T1, typename T2>
std::ostream& operator << (std::ostream& out, const Coverage<T1, T2>& r)
{
    return out << r.id;
}
template <typename T1, typename T2>
bool operator == (const typename Coverage<T1, T2>::Id& l, const Coverage<T1, T2>& r)
{
    return l == r.id;
}
template <typename T1, typename T2>
bool operator != (const typename Coverage<T1, T2>::Id& l, const Coverage<T1, T2>& r)
{
    return l != r.id;
}
template <typename T1, typename T2>
bool operator < (const typename Coverage<T1, T2>::Id& l, const Coverage<T1, T2>& r)
{
    return l < r.id;
}
template <typename T1, typename T2>
bool operator > (const typename Coverage<T1, T2>::Id& l, const Coverage<T1, T2>& r)
{
    return l > r.id;
}

template<typename T>
inline size_t card()
{
    typename T::ElementRange range = T::get();
    return range.size();
}

template<typename T>
inline typename T::ElementIterator findElement(const T& id, bool throwIfMissing = false)
{
    typedef typename T::ElementIterator I;
    for (I i = T::begin(); i != T::end(); ++i)
    {
        if (id == (*i).id) return i;
    }
    if (throwIfMissing) throw std::exception(); //"Identifier out of range");
    return T::end();
}

template<typename T>
inline typename T::ElementIterator findElement(const typename T::Id& id, bool throwIfMissing = false)
{
    typedef typename T::ElementIterator I;
    for (I i = T::begin(); i != T::end(); ++i)
    {
        if (id == (*i).id) return i;
    }
    if (throwIfMissing) throw std::exception(); //"Identifier out of range");
    return T::end();
}

template<typename T>
inline const char* findLabel(const T& id, bool throwIfMissing = false)
{
    typename T::ElementIterator i = findElement(id, throwIfMissing);
    if (i == T::end())
    {
        return 0;
    }
    else
    {
        return i->label.c_str();
    }
}

template<typename T>
inline const char* findLabel(const typename T::id& id, bool throwIfMissing = false)
{
    typename T::ElementIterator i = findElement(id, throwIfMissing);
    if (i == T::end())
    {
        return 0;
    }
    else
    {
        return i->label.c_str();
    }
}

template<typename _Id>
struct _True
{
    inline bool operator()(const _Id&) const
    {
        return true;
    }
};

template<typename _Id, typename MatchList>
struct Or
{
    inline bool operator()(const _Id& id) const
    {
        typedef typename MatchList::Predicate Predicate;
        typedef typename MatchList::PredicateIterator I;
        typedef typename MatchList::PredicateRange R;

        R r = MatchList::get();
        for (I i = r.begin(); i!= r.end(); ++i)
        {
            Predicate p = *i;
            bool check = p(id);
            if (check) return true;
        }
        return false;
    }
};

template<typename _Id, typename Predicate>
struct Not
{
    Predicate p;
    Not() : p() {}
    Not(const Predicate& p) : p(p) {}
    Not(const Not& n) : p(n.p) {}
    Not& operator= (const Not& n)
    {
        p = n.p;
        return  *this;
    }
    inline bool operator()(const _Id& id) const
    {
        return !p(id);
    }
};

template<typename _Elements, typename Predicate>
struct FilteredElements : public _Elements
{
    typedef typename _Elements::Id Id;
    typedef typename _Elements::Element Element;
    typedef typename std::vector<Element>::const_iterator ElementIterator;
    typedef std::vector<Element> ElementRange;
protected:
    static ElementRange elements;
    static bool initialize()
    {
        Predicate m;
        typename _Elements::ElementRange r = _Elements::get();
        for (typename _Elements::ElementIterator i = r.begin(); i != r.end(); ++i)
        {
            if (m((*i).id)) elements.push_back( *i );
        }
        return true;
    }
public:
    static ElementRange& get()
    {
        static bool _init = initialize();
        cpptesk::utils::_ignore(_init);
        return elements;
    }
    typedef typename _Elements::Matcher Matcher;
};

template<typename _Elements, typename Predicate>
typename FilteredElements<_Elements, Predicate>::ElementRange FilteredElements<_Elements, Predicate>::elements;

template<typename _E1, typename _E2>
struct Composition
{
    typedef _E1 Elements1;
    typedef _E2 Elements2;
    typedef typename _E1::Id Id1;
    typedef typename _E2::Id Id2;

    typedef ComposedId<Id1, Id2> Id;
    typedef CoverageElement<Id> Element;
    typedef std::vector<Element> _Storage;
    typedef typename _Storage::const_iterator ElementIterator;
    typedef _Storage ElementRange;

protected:
    static ElementRange elements;
    static bool initialize()
    {
        typename _E1::ElementRange r1 = _E1::get();
        typename _E2::ElementRange r2 = _E2::get();
        typedef typename _E1::ElementIterator I1;
        typedef typename _E2::ElementIterator I2;
        for(I1 i1 = r1.begin(); i1 != r1.end(); ++i1)
        {
            for(I2 i2 = r2.begin(); i2 != r2.end(); ++i2)
            {
                Id id((*i1).id, (*i2).id);
                std::stringstream label;
                label << (*i1).label;
                label << ",";
                label << (*i2).label;
                Element e(id, label.str());
                elements.push_back(e);
            }
        }
        return true;
    }

public:
    static ElementRange& get()
    {
        static bool _init = initialize();
        cpptesk::utils::_ignore(_init);
        return elements;
    }
    struct Matcher
    {
        typename _E1::Matcher m1;
        typename _E2::Matcher m2;
        inline bool operator()(const Id& other) const
        {
            return m1(other.id1) && m2(other.id2);
        }
        inline Matcher& operator= (const Id& id)
        {
            m1 = id.id1;
            m2 = id.id2;
            return *this;
        }
    };
};

template<typename E1, typename E2>
typename Composition<E1,E2>::ElementRange Composition<E1,E2>::elements;

}
}
}

#include <utils/macros.h>

///        .
#define CPPTESK_DEFINE_COVERAGE_NAME(_TypeName, _name)									\
struct _TypeName ##	_$_name { static inline const char * name() { return _name; }	}	\

//          .

#define CPPTESK_$_FIRST(seq,_arg) CTESK_PP_GET_0_SEQ(seq)
#define CPPTESK_$_SECOND(seq,_arg) CTESK_PP_GET_1_SEQ(seq)
#define CPPTESK_$_PUSH_ELEMENT(seq,storage)\
(storage).push_back(Element(Id(CTESK_PP_GET_0_SEQ(seq)), CTESK_PP_GET_1_SEQ(seq))) \

#define CPPTESK_$_MAKE_ELEMENT(seq,storage)\
Element(Id(CTESK_PP_GET_0_SEQ(seq)), CTESK_PP_GET_1_SEQ(seq)) \


#define CPPTESK_DEFINE_ENUMERATION(_TypeName,_elems)												\
struct _TypeName ## _$_enum {																\
	typedef enum { _$_ZERO, CTESK_PP_$_UNPAREN(CTESK_PP_TRANSFORM(CPPTESK_$_FIRST, , _elems)) } Id;				\
	typedef cpptesk::ts::coverage::CoverageElement<Id> Element;											\
	typedef Element* ElementIterator;							\
	typedef cpptesk::utils::Range<ElementIterator> ElementRange;	\
    static ElementRange get() {														\
        static Element els[] = {CTESK_PP_$_UNPAREN(CTESK_PP_TRANSFORM(CPPTESK_$_MAKE_ELEMENT, _Storage::elements, _elems))}; \
		static size_t size = sizeof(els)/sizeof(els[0]);		\
																	\
		return ElementRange(els, els+size, size);				\
	}																						\
	struct Matcher {																		\
		Id id;																				\
		inline bool operator()(const Id& other) const { return id == _$_ZERO || id == other; }	\
		inline Matcher& operator= (const Id& _id) { id = _id; return *this; }				\
	};																						\
};																							\

/// @def CPPTESK_DEFINE_ENUMERATED_COVERAGE
/// @brief    ,  .
/// @param _TypeName  ,  .
/// @param _name  ,  .
/// @param _elems     c  .  "-" 
///    ,        .
/// : <tt>(NEGATIVE, "Negative balance")</tt>
///
#define CPPTESK_DEFINE_ENUMERATED_COVERAGE(_TypeName, _name,_elems)								\
CPPTESK_DEFINE_COVERAGE_NAME(_TypeName, _name);															\
CPPTESK_DEFINE_ENUMERATION(_TypeName,_elems);													\
typedef cpptesk::ts::coverage::Coverage< _TypeName ## _$_enum, _TypeName ## _$_name> _TypeName; \

/// @def CPPTESK_DEFINE_ALIAS_COVERAGE
/// @brief   .
/// @detailed   \a _C1 c    \a Type,     \a _name.
///         ,    .
/// @param Type
/// @param _name
/// @param _C1
#define CPPTESK_DEFINE_ALIAS_COVERAGE(Type, _name, _C1)									\
CPPTESK_DEFINE_COVERAGE_NAME(Type, _name);												\
typedef _C1::Elements Type ## _$_enum;											\
typedef cpptesk::ts::coverage::Coverage< Type ## _$_enum, Type ## _$_name> Type; \

/// @def CPPTESK_DEFINE_COMPOSED_COVERAGE
/// @brief    .
/// @detailed     \a Type,     \a _name.
///        
/// <tt>ComposedId\<_C1::Id,_C2::Id\></tt>,     "%s,%s". 
///         .
///  \a _C1  \a _C2.
/// @param Type
/// @param _name
/// @param _C1
/// @param _C2
#define CPPTESK_DEFINE_COMPOSED_COVERAGE(Type, _name, _C1, _C2)							\
CPPTESK_DEFINE_COVERAGE_NAME(Type, _name);												\
typedef cpptesk::ts::coverage::Composition<_C1, _C2> Type ## _$_enum;				\
typedef cpptesk::ts::coverage::Coverage< Type ## _$_enum, Type ## _$_name> Type;\

///      .
#define CPPTESK_DEFINE_MATCH_LIST(_TypeName,_EnumType,_elems)				\
struct _TypeName ## _$_match {										\
	typedef _EnumType::Matcher Predicate;							\
	typedef Predicate* PredicateIterator;							\
	typedef cpptesk::utils::Range<PredicateIterator> PredicateRange;	\
	static inline PredicateRange get() {							\
		static Predicate preds[] = { CTESK_PP_$_UNPAREN(_elems) };			\
		static size_t size = sizeof(preds)/sizeof(preds[0]);		\
																	\
		return PredicateRange(preds, preds+size, size);				\
	}																\
};																	\

/// @def CPPTESK_DEFINE_COMPOSED_COVERAGE_EXCLUDING
/// @brief     c .
/// @detailed     \a Type,     \a _name.
///        
/// <tt>ComposedId\<_C1::Id,_C2::Id\></tt>,     "%s,%s". 
///         \a _excluded
///  .        ,
///    : <tt>({C1::A, C2::X}, {C1::B, C2::Y})</tt>.
///
///     .   
///  \c C1  \ 2,   <tt>{C1::A, C2::ANY()}</tt>   
/// ,        \c C1::A.
///
///  ,     ,  
///  .  <tt>C1 = C1_1 * C1_2</tt>, <tt>C2 = C2_1 * C2_2</tt>,
///        
/// <tt>{C1_1::A, C1_2::B, C2_1::X, C2_2::Y}</tt>.     
/// ANY    -         .
///    <tt>C1::ANY()</tt>  <tt>C2::ANY()</tt> -  
///     . , <tt>{C1::ANY(), C2_1::X, C2_2::Y}</tt>
///     ,       \c C2_1::X,   
/// \c C2_2::Y
/// @param Type
/// @param _name
/// @param _C1
/// @param _C2
#define CPPTESK_DEFINE_COMPOSED_COVERAGE_EXCLUDING(Type,_name,_C1,_C2,_excluded)					\
CPPTESK_DEFINE_COVERAGE_NAME(Type, _name);															\
typedef cpptesk::ts::coverage::Composition<_C1, _C2> Type ## _$_base_enum;					\
CPPTESK_DEFINE_MATCH_LIST(Type ## _exclude, Type ## _$_base_enum,_excluded)							\
typedef cpptesk::ts::coverage::FilteredElements<Type ## _$_base_enum,							\
	cpptesk::ts::coverage::Not<Type ## _$_base_enum::Id,										\
		cpptesk::ts::coverage::Or<Type ## _$_base_enum::Id, Type ## _exclude_$_match> > >		\
	Type ## _$_enum;																		\
typedef cpptesk::ts::coverage::Coverage< Type ## _$_enum, Type ## _$_name> Type;\

// For backward compatibility
///@deprecated
#define CPPTESK_DEFINE_COMPOSED_COVERAGE_EXLUDING CPPTESK_DEFINE_COMPOSED_COVERAGE_EXCLUDING

#if (defined(CPPTESK_SHORT_NAMES) && !defined(CPPTESK_COVERAGE_NO_SHORT_NAMES))\
    || defined(CPPTESK_COVERAGE_SHORT_NAMES)

#define DEFINE_ENUMERATED_COVERAGE CPPTESK_DEFINE_ENUMERATED_COVERAGE
#define DEFINE_ALIAS_COVERAGE CPPTESK_DEFINE_ALIAS_COVERAGE
#define DEFINE_COMPOSED_COVERAGE CPPTESK_DEFINE_COMPOSED_COVERAGE
#define DEFINE_COMPOSED_COVERAGE_EXCLUDING CPPTESK_DEFINE_COMPOSED_COVERAGE_EXCLUDING

#endif

#if (defined(CPPTESK_SHORT_SHORT_NAMES) && !defined(CPPTESK_COVERAGE_NO_SHORT_NAMES)) || defined(CPPTESK_COVERAGE_SHORT_SHORT_NAMES)

#define ENUM_COV CPPTESK_DEFINE_ENUMERATED_COVERAGE
#define ALIAS_COV CPPTESK_DEFINE_ALIAS_COVERAGE
#define COMP_COV CPPTESK_DEFINE_COMPOSED_COVERAGE
#define COMP_COV_EXCL CPPTESK_DEFINE_COMPOSED_COVERAGE_EXCLUDING

#endif

