/*
 * Copyright 2012-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK (cpptesk::examples)
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Family of models for the MOESI cache coherence protocols:
//           (1) MSI   (Modified,                   Shared, Invalid)
//           (2) MOSI  (Modified, Owned,            Shared, Invalid)
//           (3) MESI  (Modified,        Exclusive, Shared, Invalid)
//           (4) MOESI (Modified, Owned, Exclusive, Shared, Invalid)
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <moesi/processor.hpp>
#include <moesi/protocol.hpp>

#include <ts/coverage.hpp>
#include <tracer/tracer.hpp>

using namespace cpptesk::hw;
using namespace cpptesk::ts::coverage;
using namespace cpptesk::tracer::coverage;

namespace cpptesk {
namespace examples {
namespace moesi {

/**
 * Test wrapper for MOESI protocol models
 */
class MoesiModel: public Processor
{
public:
    ////////////////////////////////////////////////////////////////////////////
    // Constructor / destructor
    ////////////////////////////////////////////////////////////////////////////

    MoesiModel(CacheProtocol protocol, size_t cores_number):
        Processor(protocol, cores_number) {}

    virtual ~MoesiModel() {}

    ////////////////////////////////////////////////////////////////////////////
    // Test coverage
    ////////////////////////////////////////////////////////////////////////////

    ENUM_COV(StateCoverage, "State Coverage",
    (
        (MODIFIED,  "Modified"),
        (OWNED,     "Owned"),
        (EXCLUSIVE, "Exclusive"),
        (SHARED,    "Shared"),
        (INVALID,   "Invalid")
    ));

    StateCoverage state_coverage(CacheState state) const
    {
        switch(state)
        {
        case MODIFIED:  { return StateCoverage::MODIFIED;  }
        case OWNED:     { return StateCoverage::OWNED;     }
        case EXCLUSIVE: { return StateCoverage::EXCLUSIVE; }
        case SHARED:    { return StateCoverage::SHARED;    }
        default:        { return StateCoverage::INVALID;   }
        }
    }

    ////////////////////////////////////////////////////////////////////////////

    ENUM_COV(StimulusCoverage, "Stimulus Coverage",
    (
        (READ,  "Read"),
        (WRITE, "Write"),
        (EVICT, "Evict")
    ));

    StimulusCoverage stimulus_coverage(CacheStimulus stimulus) const
    {
        switch(stimulus)
        {
        case READ:  { return StimulusCoverage::READ;  }
        case WRITE: { return StimulusCoverage::WRITE; }
        default:    { return StimulusCoverage::EVICT; }
        }
    }

    ////////////////////////////////////////////////////////////////////////////

    COMP_COV(CacheCoverage, "Cache Coverage", StateCoverage, StimulusCoverage);

    CacheCoverage cache_coverage(CacheState state, CacheStimulus stimulus) const
        { return state_coverage(state) * stimulus_coverage(stimulus); }

    ////////////////////////////////////////////////////////////////////////////

    void track_coverage(CacheState state, CacheStimulus stimulus)
    {
        _coverage_tracker << cache_coverage(state, stimulus);
    }

    ////////////////////////////////////////////////////////////////////////////
    // Operations under test
    ////////////////////////////////////////////////////////////////////////////

    /// Emulates the read operation on the i-th core.
    void read(size_t i)  { exec(is_initial_state() ? READX : READ, i); }
    /// Emulates the write operation on the i-th core.
    void write(size_t i) { exec(WRITE, i); }
    /// Emulates the eviction operation on the i-th core.
    void evict(size_t i) { exec(EVICT, i); }

    ////////////////////////////////////////////////////////////////////////////
    // Cache coherence property
    ////////////////////////////////////////////////////////////////////////////

    /// Cache coherence property (invariant of any cache coherence protocol).
    virtual bool invariant()
    {
        size_t M = 0;
        size_t O = 0;
        size_t E = 0;
        size_t S = 0;

        // Collect information on all processor cores.
        for(size_t i = 0; i < get_cores_number(); i++)
        {
            switch(get_core(i).get_state())
            {
            case MODIFIED:  { M++; break; }
            case OWNED:     { O++; break; }
            case EXCLUSIVE: { E++; break; }
            case SHARED:    { S++; break; }
            default:        {    ; break; }
            }
        }

        // The mutually exclusive combinations are as follows:
        // M&M, O&O, E&E, M&O, M&E, M&S, O&E and E&S.
        return (M <= 1) && !(M && O) && !(M && E) && !(M && S)
            && (O <= 1) && !(O && E)
            && (E <= 1) && !(E && S);
    }

private:
    /// Emulates the operation on the i-th core.
    void exec(CacheStimulus stimulus, size_t i)
    {
        if(pre_recv(stimulus, i))
        {
            track_coverage(get_core(i).get_state(), stimulus);
            recv(stimulus, i);
        }
    }

private:
    /// UTT2 coverage tracker.
    CoverageTracker _coverage_tracker;
};

}}} // cpptesk::examples::moesi

