/** @file fifo_media.cpp
 * Mediator for united FIFO model.
 *
 * @author Mikhail Chupilko (chupilko@ispras.ru)
 */

#include <multififo_media.h>

#include <sync.h>

#include <interface.h>
#include <vpi_media.h>

#include <iostream>

namespace cpptesk {
namespace examples {
namespace fifo {

MultififoMediator::MultififoMediator()
{
    // Set level of debug messages
    CPPTESK_SET_DEBUG_LEVEL(cpptesk::hw::DEBUG_USER, true);
    CPPTESK_SET_DEBUG_STYLE(cpptesk::hw::DEBUG_USER, cpptesk::hw::BLACK, cpptesk::hw::CYAN, cpptesk::hw::CYAN, cpptesk::hw::BLACK);

    // Registration of correspondence between serializers, deserializers and ifaces.
    CPPTESK_SET_INPUT_ADAPTER(iface1a, MultififoMediator::serialize_iface1a);
    CPPTESK_SET_INPUT_ADAPTER(iface1b, MultififoMediator::serialize_iface1b);
    CPPTESK_SET_INPUT_ADAPTER(iface2, MultififoMediator::serialize_iface2);
    CPPTESK_SET_OUTPUT_ADAPTER_NO_LISTENER(iface3, MultififoMediator::deserialize_iface3);
//    CPPTESK_SET_ARBITER(iface3, matcher_iface3);

    CPPTESK_SET_OUTPUT_ADAPTER_OF(FIFOMediator, fifo0, fifo0.iface3, MultififoMediator::deserialize_inner_iface0);
    CPPTESK_SET_OUTPUT_ADAPTER_OF(FIFOMediator, fifo1, fifo1.iface3, MultififoMediator::deserialize_inner_iface1);

    // Set maximum timeout.
    CPPTESK_SET_REACTION_TIMEOUT(MultififoMediator::REACTION_TIMEOUT);
}

MultififoMediator::~MultififoMediator() {}

bool MultififoMediator::isIface1aReady() const
{
    return Multififo::isIface1aReady();
}
bool MultififoMediator::isIface1bReady() const
{
    return Multififo::isIface1bReady();
}
bool MultififoMediator::isIface2Ready() const
{
    return Multififo::isIface2Ready();
}

CPPTESK_DEFINE_PROCESS(MultififoMediator::serialize_iface1a)
{
    // Copy data to the local variable.
    InputData msg = CPPTESK_CAST_MESSAGE(InputData);

    // Start of serialization processing.
    CPPTESK_START_PROCESS();

    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_USER, "MultififoMediator::serialize_iface1a " << msg);

    // Capture iface1a.
    CPPTESK_CAPTURE_IFACE();

    // Set data to the wires of DUT.
    inputs.CE_WR0 = 1;
    inputs.DI0 = msg.get_data();

    // Release of iface1a.
    CPPTESK_RELEASE_IFACE();

    // Wait one cycle of simulation.
    CPPTESK_CYCLE();

    // Finish of serialization processing.
    CPPTESK_STOP_PROCESS();
}

CPPTESK_DEFINE_PROCESS(MultififoMediator::serialize_iface1b)
{
    // Copy data to the local variable.
    InputData msg = CPPTESK_CAST_MESSAGE(InputData);

    // Start of serialization processing.
    CPPTESK_START_PROCESS();

    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_USER, "MultififoMediator::serialize_iface1b " << msg);

    // Capture iface1b.
    CPPTESK_CAPTURE_IFACE();

    // Set data to the wires of DUT.
    inputs.CE_WR1 = 1;
    inputs.DI1 = msg.get_data();

    // Release of iface1b.
    CPPTESK_RELEASE_IFACE();

    // Wait one cycle of simulation.
    CPPTESK_CYCLE();

    // Finish of serialization processing.
    CPPTESK_STOP_PROCESS();
}

CPPTESK_DEFINE_PROCESS(MultififoMediator::serialize_iface2)
{
    // Copy data to the local variable.
    InputData msg = CPPTESK_CAST_MESSAGE(InputData);

    // Start of serialization processing.
    CPPTESK_START_PROCESS();

    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_USER, "MultififoMediator::serialize_iface2 " << msg);

    // Capture iface2.
    CPPTESK_CAPTURE_IFACE();

    // Set data to the wires of DUT.
    inputs.CE_RD = 1;

    // Release of iface2.
    CPPTESK_RELEASE_IFACE();

    // Wait one cycle of simulation.
    CPPTESK_CYCLE();

    // Finish of serialization processing.
    CPPTESK_STOP_PROCESS();
}

CPPTESK_DEFINE_PROCESS(MultififoMediator::deserialize_iface3)
{
    // Copy data to the local variable.
    OutputData &msg = CPPTESK_CAST_MESSAGE(OutputData);

    // Start of deserialization processing.
    CPPTESK_START_PROCESS();

    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_USER, "MultififoMediator::deserialize_iface3 " << msg.forceToString());

    // Wainting for reaction by means of DUT's DO_VAL wire.
    CPPTESK_WAIT_REACTION(outputs.DO_VAL);

    // Output data are saved to the following check.
    msg.set_data(outputs.DO);

    // Jump to the next registered reaction (if it exists).
    CPPTESK_NEXT_REACTION();

    // Finish of deserialization processing.
    CPPTESK_STOP_PROCESS();
}

CPPTESK_DEFINE_PROCESS(MultififoMediator::deserialize_inner_iface0)
{
    OutputData msg = CPPTESK_CAST_MESSAGE(OutputData);

    CPPTESK_START_PROCESS();

    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_USER, "MultififoMediator::deserialize_inner_iface0 " << msg.forceToString());

    CPPTESK_CAPTURE_IFACE();

    if(!fifo2.is_full()) {
        InputData data = InputData();
        data.set_data(msg.get_data());
        CPPTESK_CALL_STIMULUS_OF(fifo2, FIFO::push_msg, fifo2.iface1, data);
    }
    else
    {
        CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_USER, "but the fifo2 is full " << msg.forceToString());
    }

    CPPTESK_RELEASE_IFACE();

    CPPTESK_CYCLE_OF(fifo0);

    CPPTESK_STOP_PROCESS();
}

CPPTESK_DEFINE_PROCESS(MultififoMediator::deserialize_inner_iface1)
{
    OutputData msg = CPPTESK_CAST_MESSAGE(OutputData);

    CPPTESK_START_PROCESS();

    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_USER, "MultififoMediator::deserialize_inner_iface1 " << msg.forceToString());

    while(!fifo2.isIface1Ready()) CPPTESK_CYCLE_OF(fifo1);

    if(!fifo2.is_full()) {
	CPPTESK_CAPTURE_IFACE();
        InputData data = InputData();
        data.set_data(msg.get_data());
        CPPTESK_CALL_STIMULUS_OF(fifo2, FIFO::push_msg, fifo2.iface1, data);
	CPPTESK_RELEASE_IFACE();
    }
    else
    {
        CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_USER, "but the fifo2 is full " << msg.forceToString());
    }

    CPPTESK_CYCLE_OF(fifo1);

    CPPTESK_STOP_PROCESS();
}

void MultififoMediator::cycle()
{
    fifo0.cycle();
    fifo1.cycle();
    fifo2.cycle();
    Multififo::cycle();
}

}}}
