/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Information Metric for rules
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/metrics/information_metric.hpp>
#include <hw/diagnostics/metrics/field_metric.hpp> //Line: 47

#include <map>
#include <math.h>

namespace cpptesk {
namespace hw {
namespace diagnostics {

double InformationMetric::metric_threshold = 0.45;

void InformationMetric::set_min_threshold(const double threshold)
{
        metric_threshold = threshold;
}

InformationMetric* InformationMetric::getMetric()
{
    return cpptesk::hw::diagnostics::SingleMetric<InformationMetric>::getInstance();
}
    
/// Return metric of elements
int InformationMetric::compare(const cpptesk::hw::Message &m1, const cpptesk::hw::Message &m2)
{
    if(FieldMetric::fields_metric(m1, m2) == EXACT_MATCH)
        return EXACT_MATCH;

    if(m1.size() == m2.size())
    {
            
        InformationMetric::DataMessage<uint64_t> data_message1 = get_data<uint64_t>(m1);
        InformationMetric::DataMessage<uint64_t> data_message2 = get_data<uint64_t>(m2);
        
        MessageMetric message_metric;
        MessageMetric message_metric2;
        
        for(int i = 0; i < (int)data_message1.data.size(); i++)
        {
            // Message 1
            InformationMetric::DataList<uint64_t> data_list  = data_message1.data[i];

            double res_x = estimate_information<uint64_t>(data_list);
            //CPPTESK_DEBUG_STREAM() << "Res(x) = [" << res_x << "]" << std::endl;

            // Message 2
            InformationMetric::DataList<uint64_t> data_list2 = data_message2.data[i];
            
            double res_y = estimate_information<uint64_t>(data_list2);
            //CPPTESK_DEBUG_STREAM() << "Res(y) = [" << res_y << "]" << std::endl;

            // Metric
            double res_temp_1 = estimate_information2<uint64_t>(data_list, data_list2);
            //CPPTESK_DEBUG_STREAM()  <<  "Res4 = [" << res_temp_1 << "]" << std::endl;

            double result_1;
            if (res_x == 0) result_1 = 0;
            else result_1 = res_temp_1/res_x;

            //CPPTESK_DEBUG_STREAM() << "Result{res_temp_1/res_x} = [" << result_1 << "]" << std::endl;
            
            double res_temp_2 = estimate_information2<uint64_t>(data_list2, data_list);
            //CPPTESK_DEBUG_STREAM()  <<  "Res4 = [" << res_temp_2 << "]" << std::endl;

            double result_2;
            if (res_y == 0) result_2 = 0;
            else result_2 = res_temp_2/res_y;

            //CPPTESK_DEBUG_STREAM() << "Result{res_temp_2/res_y} = [" << result_2 << "]" << std::endl;

            double result = (result_1 + result_2)/2;
            
            //CPPTESK_DEBUG_STREAM() << "Result = [" << result << "]" << std::endl;        
            
            //CPPTESK_DEBUG_STREAM()  << std::endl << "--------------------------------------" << std::endl;
            
            message_metric.add(res_x, res_y, result);
        }
    
        double result_m = 0.0;
        double result_w = 0.0;
        for(int i = 0; i < message_metric.size(); i++)
        {
            double w = (message_metric.unit[i].i_x + message_metric.unit[i].i_y)/2;
            
            if(w > 1.0) w = 1.0;
            
            result_m += (w*message_metric.unit[i].field_metric);
            result_w += w;
        }
        
        result_m = WORST_MATCH*result_m/result_w;
        
        //CPPTESK_DEBUG_STREAM()  << "Result = " << result_m << std::endl;
        //CPPTESK_DEBUG_STREAM()  << std::endl << "=====================================" << std::endl << std::endl << std::endl;

        if(result_m > (double)WORST_MATCH*metric_threshold) result_m = WORST_MATCH;
        
        //CPPTESK_DEBUG_STREAM()  << std::endl << (double)WORST_MATCH*metric_threshold << "<" << result_m << std::endl;
    
        return (int)result_m;
    }
    else
        return FALSE_MATCH;
}

template <typename T>
double InformationMetric::estimate_information2(InformationMetric::DataList<T> data_list, InformationMetric::DataList<T> data_list2)
{
    int n = data_list.size();
    //CPPTESK_DEBUG_STREAM() << "Elements(" << std::dec << data_list.data.size() << "): ";
    
    double i_x_j = 0.0;
        
    std::map <uint64_t, int> map_msg;

    int common_element = 0;
    
    for(int j = 0; j < (int)data_list.data.size(); j++)
    {
        //bool temp_bool = false;
        if(data_list.data[data_list.data.size() - j - 1].get_data() == data_list2.data[data_list.data.size() - j - 1].get_data())
        {
            common_element++;
            
            //temp_bool = true;
        }
        else
            map_msg[data_list.data[data_list.data.size() - j - 1].get_data()]++;

        // Print for debug
        /*CPPTESK_DEBUG_STREAM() << "{" << CPPTESK_DEBUG_STYLE(GREEN, temp_bool?BLACK:YELLOW) << "[" <<
            std::hex << data_list.data[data_list.data.size() - j - 1].get_data() << 
            "]&" << CPPTESK_DEBUG_STYLE(BLUE, temp_bool?BLACK:YELLOW) << "[" << std::hex << 
            data_list2.data[data_list2.data.size() - j - 1].get_data() << "]" << 
            CPPTESK_DEBUG_STYLE(NORMAL, NORMAL) << "}";*/
    }
    
    //CPPTESK_DEBUG_STREAM() << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL) << std::endl;

    // Get most common element
    std::map<uint64_t,int>::iterator common = map_msg.begin();
    int max_common = common->first;
    
    for(std::map<uint64_t,int>::iterator it=map_msg.begin(); it!=map_msg.end(); ++it)
        if(map_msg[max_common] < it->second)
        {
            max_common = it->first;
        }

    map_msg[max_common] += common_element;
    
    for(std::map<uint64_t,int>::iterator it=map_msg.begin(); it!=map_msg.end(); ++it)
    {
        i_x_j = i_x_j + it->second*(log2((double)n/(double)it->second));
        //CPPTESK_DEBUG_STREAM() << "[" << it->first << " => " << it->second << "]";
    }
    //CPPTESK_DEBUG_STREAM() << std::endl;
    
    return i_x_j;
}

// I(part(X|Y))
template <typename T>
double InformationMetric::segments_partitions(InformationMetric::DataList<T> data_list1, InformationMetric::DataList<T> data_list2)
{
    std::map <uint64_t, int> map_msg;
    
    int count = 1;
    bool segment, new_segment;    

    if(data_list1.data[0].get_data() == data_list2.data[0].get_data())
        segment = true;
    else 
        segment = false;

    for(int i = 0; i < (int)data_list1.data.size(); i++)
    {
        if(data_list1.data[i].get_data() == data_list2.data[i].get_data())
            new_segment = true;
        else
            new_segment = false;
        
        if(new_segment != segment)
        {
            count++;
            map_msg[count]++;
            segment = new_segment;
        }
        else
            map_msg[count]++;
    }
    
    std::map <uint64_t, int> map_msg2;
    for(std::map<uint64_t,int>::iterator it=map_msg.begin(); it!=map_msg.end(); ++it)
    {
        //CPPTESK_DEBUG_STREAM() << "[" << it->first << " => " << it->second << "]";
        map_msg2[it->second] ++;
    }
    //CPPTESK_DEBUG_STREAM() << std::endl;
    
    int sum = map_msg.size();

    double result = 0.0;

    for(std::map<uint64_t,int>::iterator it=map_msg2.begin(); it!=map_msg2.end(); ++it)
    {
        //CPPTESK_DEBUG_STREAM() << "[" << it->first << " => " << it->second << "]";
        
        if(((double)(sum - 1)/it->second > 1) && (it->second != 0))
            result = (double)sum * log2((double)(sum - 1)/it->second);
    }
    //CPPTESK_DEBUG_STREAM() << std::endl;

    if(result <= 1.0)  return 0;
    
    return result;
}

// I(diff(X|Y))
template <typename T>
double InformationMetric::estimate_dif(InformationMetric::DataList<T> data_list1, InformationMetric::DataList<T> data_list2)
{
    InformationMetric::DataList<T> data_list;
    
    for(int j = 0; j < (int)data_list1.data.size(); j++)
    {
        if(data_list1.data[j].get_data() != data_list2.data[j].get_data())
            data_list.push_back(data_list1.data[j]);
    }
    
    return estimate_information(data_list);
}

template <typename T>
double InformationMetric::estimate_information(InformationMetric::DataList<T> data_list)
{
    int n = data_list.size();
    //CPPTESK_DEBUG_STREAM() << "Elements(" << std::dec << data_list.data.size() << "): ";
    
    double i_x_j = 0.0;
        
    std::map <T, int> map_msg;
    
    for(int j = 0; j < (int)data_list.data.size(); j++)
    {
        map_msg[data_list.data[data_list.data.size() - j - 1].get_data()]++;

        //CPPTESK_DEBUG_STREAM() << "[" << std::hex << data_list.data[data_list.data.size() - j - 1].get_data() << "]";
    }
    //CPPTESK_DEBUG_STREAM() << std::endl;
    
    for(std::map<uint64_t,int>::iterator it=map_msg.begin(); it!=map_msg.end(); ++it)
    {
        if(((double)n/(double)it->second > 1) && ((double)it->second != 0))
            i_x_j = i_x_j + it->second*(log2((double)n/(double)it->second));
        
        //CPPTESK_DEBUG_STREAM() << "[" << it->first << " => " << it->second << "]";
    }
    //CPPTESK_DEBUG_STREAM() << std::endl;
    
    return i_x_j;
}

/// Transformation Message into DataMessage
template <typename T>
InformationMetric::DataMessage<T> InformationMetric::get_data(const cpptesk::hw::Message &m1)
{
    InformationMetric::DataMessage<T> data_message;

    for(cpptesk::hw::Message::const_iterator i = m1.begin(); i != m1.end(); i++)
    {
        InformationMetric::DataList<T> data_list;
    
        const Field<> &field = m1.**i;

        if(field.isComparable())
        {
            data_list.clear();
            
            // Basic item length
            int item_length = get_item_length(field.size() * field.get_width());
            
            for(unsigned j = 0; j < field.size(); j++)
            {
                Data<T> data_item;

                for(unsigned k = 0; k < field.get_width(); k++)
                {
                    data_item.push_back(field.getBit(j*field.get_width() + k));
                    
                    if((k+1)%item_length == 0 || 
                       ((int)field.get_width() < item_length && (k+1)%field.get_width() == 0) ||
                       data_item.size() == item_length || (k+1 == field.get_width() && j+1 == field.size()))
                    {
                        data_list.push_back(data_item);
                        data_item.clear();
                    }
                }
            }
            data_message.push_back(data_list);
        }
    }

    return data_message;
}

/// Return item length
//0  <= n <= 15 -> length 1
//16 <= n <= 31 -> length 2
//32 <= n <= 63 -> length 4
//64 <= n <= ++ -> length 8
int InformationMetric::get_item_length(const int n)
{
    if(n <= 15) return 1;
    if(n <= 31) return 2;
    if(n <= 63) return 4;
    if(n >= 64) return 8;
    
    return 8;
}
}}}
