/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Information Metric for rules
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <vector>

#include <hw/diagnostics/rules/rule.hpp>
#include <hw/diagnostics/metrics/metric.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {
    
class InformationMetric: public BaseMetric
{
    friend class SingleMetric<InformationMetric>;

private:
    /**
    * The minimum element structure
    */
    template <typename T>
    struct Data
    {
        std::vector<T> data;

        void push_back(T item_data) { this->data.push_back(item_data); }
        void clear() {this->data.clear();};
        int size() {return (int)this->data.size();};
        
        size_t lo_index;
        size_t hi_index;

        T get_data()
        {
            T data_value = 0x0;
            
            for(int i = 0; i < (int)this->data.size(); i++)
            {
                data_value <<= 1;
                data_value |= this->data[this->data.size() - i - 1];
            }
            
            return data_value;
        }

        std::string to_str()
        {
            std::string temp;
            std::stringstream temp_stream;
            
            temp_stream << std::hex << this->get_data();
            temp.append(temp_stream.str());
            
            return temp;
        }
    };
    
    template <typename T>
    struct DataList: public Data< Data<T> >
    {
    };

    template <typename T>
    struct DataMessage: public Data< DataList<T> >
    {};

    struct MessageUnit
    {
        double i_x;
        double i_y;
        double field_metric;
    };

    struct MessageMetric
    {
        std::vector<MessageUnit> unit;

        int size() {return (int)this->unit.size();};

        void add(double x, double y, double result)
        {
            MessageUnit message_unit;
            message_unit.i_x = x;
            message_unit.i_y = y;
            message_unit.field_metric = result;
            add(message_unit);
        }
        
        void add(MessageUnit result)
        {
            unit.push_back(result);
        }
    };

protected:
    /// Constructor for Metric class
    InformationMetric() {}
    /// Destructor for Metric class
    virtual ~InformationMetric() {}

public:
    /// Return metric of elements
    int compare(const cpptesk::hw::Message &m1, const cpptesk::hw::Message &m2);

    /// Return metric synopsis
    std::string toString() {return "Information Metric";}    
    std::string operator<< (InformationMetric &metric) {return metric.toString();}
    
    static InformationMetric* getMetric();
private:
    
    /// Minimum value to use the rule
    static void set_min_threshold(const double threshold);
    static double metric_threshold;
    
    template <typename T>
    static InformationMetric::DataMessage<T> get_data(const cpptesk::hw::Message &m1);
    
    template <typename T>
    static double estimate_information2(InformationMetric::DataList<T> data_list, InformationMetric::DataList<T> data_list2);

    template <typename T>
    static double estimate_information(InformationMetric::DataList<T> data_list);
    
    template <typename T>
    static double estimate_dif(InformationMetric::DataList<T> data_list1, InformationMetric::DataList<T> data_list2);
    
    template <typename T>
    static double segments_partitions(InformationMetric::DataList<T> data_list1, InformationMetric::DataList<T> data_list2);
    
    /// Return item length
    static int get_item_length(const int n);
}; 

    
}}}
