/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: ReactionUnit class keeps Reaction and IDs to its predecessors
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/reaction_unit.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

ReactionUnit::ReactionUnit() {}
ReactionUnit::~ReactionUnit() {}

ReactionUnit::ReactionUnit(const ReactionUnit &ru): _reaction(ru._reaction)
{
    _parent_spec_id  = ru._parent_spec_id;
    _parent_impl_id  = ru._parent_impl_id;
    _collapsed       = ru._collapsed;
    _id              = ru._id;
    _collapse_type   = ru._collapse_type;
    _collapse_rule   = ru._collapse_rule;
}

ReactionUnit::ReactionUnit(const cpptesk::hw::Reaction &r) : _reaction(r)
{
    _parent_spec_id  = -1;
    _parent_impl_id  = -1;
    _collapsed       = false;
    _id              = -1;
    _collapse_type   = NO_COLLAPSE;
    _collapse_rule   = -1;
}

ReactionUnit::ReactionUnit(const cpptesk::hw::Reaction &r, int parent_spec_id, int parent_impl_id) : _reaction(r)
{
    _parent_spec_id  = parent_spec_id;
    _parent_impl_id  = parent_impl_id;
    _collapsed       = false;
    _id              = -1;
    _collapse_type   = NO_COLLAPSE;
    _collapse_rule   = -1;
}

void ReactionUnit::set_collapsed()
{
    _collapsed = true;
}

bool ReactionUnit::is_collapsed()
{
    return _collapsed;
}

void ReactionUnit::set_collapse_rule(int rule)
{
    _collapse_rule = rule;
}

int ReactionUnit::get_collapse_rule()
{
    return _collapse_rule;
}

void ReactionUnit::set_collapse_type(int collapse_type)
{
    _collapse_type = collapse_type;
}

int ReactionUnit::get_collapse_type()
{
    return _collapse_type;
}

void ReactionUnit::set_id(int id)
{
    _id = id;
}

int ReactionUnit::get_id()
{
    return _id;
}

void ReactionUnit::set_parent_spec_id(int parent_id)
{
    _parent_spec_id = parent_id;
}

int ReactionUnit::get_parent_spec_id()
{
    return _parent_spec_id;
}

void ReactionUnit::set_parent_impl_id(int parent_id)
{
    _parent_impl_id = parent_id;
}

int ReactionUnit::get_parent_impl_id()
{
    return _parent_impl_id;
}

void ReactionUnit::print_description(dbg_level_t dbg_level, dbg_color_t fg_level, dbg_color_t bg_level)
{
    CPPTESK_COLORED_DEBUG_PRINT(dbg_level, fg_level, bg_level, toString());
}

void ReactionUnit::print_description(dbg_level_t dbg_level)
{
    print_description(dbg_level, cpptesk::hw::DEFAULT, cpptesk::hw::DEFAULT);
}

void ReactionUnit::print_description()
{
    print_description(cpptesk::hw::DEBUG_BASIC, cpptesk::hw::DEFAULT, cpptesk::hw::DEFAULT);
}

void ReactionUnit::print_short_description(dbg_level_t dbg_level, dbg_color_t fg_level, dbg_color_t bg_level, bool spec)
{
    if(spec)
    {
        CPPTESK_COLORED_DEBUG_PRINT(dbg_level, fg_level, bg_level,
             _reaction.getSpecMessage()->forceToString());
    }
    else
    {
        CPPTESK_COLORED_DEBUG_PRINT(dbg_level, fg_level, bg_level,
             _reaction.getImplMessage()->forceToString());
    }
}

void ReactionUnit::print_short_description(dbg_level_t dbg_level, bool spec)
{
    print_short_description(dbg_level, cpptesk::hw::DEFAULT, cpptesk::hw::DEFAULT, spec);
}

std::string ReactionUnit::operator<< (ReactionUnit &ru)
{
    return ru.toString();
}

std::string ReactionUnit::toString()
{
    std::stringstream stream;
    if(_reaction.getError() == MISSING_REACTION)
    {
        stream << _reaction.getShortDescription() << std::endl <<
                  " SpecMessage: " << _reaction.getSpecMessage()->forceToString();
    }
    else if(_reaction.getError() == UNEXPECTED_REACTION)
    {
        stream << _reaction.getShortDescription() << std::endl <<
                  " ImplMessage: " << _reaction.getImplMessage()->forceToString();
    }
    else
    {
        stream << _reaction.getShortDescription() << std::endl <<
                  " SpecMessage: " << _reaction.getSpecMessage()->forceToString() << std::endl <<
                  " ImplMessage: " << _reaction.getImplMessage()->forceToString();
    }
    return stream.str();
}

cpptesk::hw::Reaction& ReactionUnit::getReaction()
{
    return _reaction;
}

}}} // cpptesk::hw::diagnostics
