/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Combining MISSING and UNEXPECTED reaction pairs
//           Rule #3: {[a:x]+[x:b] => [a:b]+[x:x]}
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/rules/rule3.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

int Rule3::check(cpptesk::hw::Reaction &r1, cpptesk::hw::Reaction &r2)
{
    if((r1.getError() == MISSING_REACTION && r2.getError() == UNEXPECTED_REACTION &&
       r1.getSpecMessage() != NULL        && r2.getImplMessage() != NULL &&
       r1.getSpecMessage()->getName()     == r2.getImplMessage()->getName()) ||
      (r2.getError() == MISSING_REACTION && r1.getError() == UNEXPECTED_REACTION &&
       r2.getSpecMessage() != NULL        && r1.getImplMessage() != NULL &&
       r2.getSpecMessage()->getName()     == r1.getImplMessage()->getName()))
    {
        cpptesk::hw::Reaction &in_r1 = (r1.getError() == MISSING_REACTION ? r1 : r2);
        cpptesk::hw::Reaction &in_r2 = (r1.getError() == MISSING_REACTION ? r2 : r1);

        int metric_ab = compare(in_r1.getSpecMessage(), in_r2.getImplMessage());

        return metric_ab;
    }

    return FALSE_MATCH;
}

ReactionUnitVector Rule3::apply(ReactionUnit &ru1, ReactionUnit &ru2)
{
    cpptesk::hw::Reaction &r1 = ru1.getReaction();
    cpptesk::hw::Reaction &r2 = ru2.getReaction();

    bool is_r1_missing = r1.getError() == MISSING_REACTION;

    const cpptesk::hw::Message &m1 = *(is_r1_missing ? r1.getSpecMessage() : r2.getSpecMessage());
    const cpptesk::hw::Message &m2 = *(is_r1_missing ? r2.getImplMessage() : r1.getImplMessage());

    int id1 = (is_r1_missing ? ru1.get_id() : ru2.get_id());
    int id2 = (is_r1_missing ? ru2.get_id() : ru1.get_id());

    // Take min "start" and max "end" time of reactions
    cycle_t start_time = r1.whenStarted() < r2.whenStarted() ? r1.whenStarted() : r2.whenStarted();
    cycle_t end_time   = r1.whenStopped() > r2.whenStopped() ? r1.whenStopped() : r2.whenStopped();

    bool normal = (compare(m1, m2) == EXACT_MATCH);
    cpptesk::hw::Reaction new_r1 = Reaction(normal ? NORMAL_REACTION : INCORRECT_REACTION, *(r1.where()), m1, m2,
                                            start_time, end_time, normal ? "normal reaction after rule#3" :
                                                                           "incorrect reaction after rule#3");
    CPPTESK_ASSERTION(r1.where() != NULL, "Interface should not be NULL");

    ReactionUnitVector result = ReactionUnitVector(new_r1, id1, id2); // (reaction, spec_parent, impl_parent)

    ru1.set_collapsed();
    ru2.set_collapsed();

    return result;
}

std::string Rule3::toString()
{
    return "for MISSING and UNEXPECTED reactions: {[a:x]+[x:b] => [a:b]+[x:x]}";
}

}}}
