/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Message field
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/random.hpp>

#include <stdint.h>

#include <sstream>
#include <ostream>
#include <string>
#include <vector>

namespace cpptesk {
namespace hw {

inline uint16_t reverse_byte_order(uint16_t value)
{
    return ((value & 0xff) << 8)
         | ((value >> 8) & 0xff);
}

inline uint32_t reverse_byte_order(uint32_t value)
{
    return (uint32_t)reverse_byte_order((uint16_t)((value & 0xffff))) << 16
         | (uint32_t)reverse_byte_order((uint16_t)((value >> 16) & 0xffff));
}

inline uint64_t reverse_byte_order(uint64_t value)
{
    return (uint64_t)reverse_byte_order((uint32_t)((value & 0xffffFFFFul))) << 32
         | (uint64_t)reverse_byte_order((uint32_t)((value >> 32) & 0xffffFFFFul));
}

inline uint64_t get_mask(unsigned width)
{
    return width == 64 ? 0xffffFFFFffffFFFFull : ((1ull << width) - 1);
}

inline uint64_t get_mask(unsigned min, unsigned max)
{
    if(min > max) { unsigned tmp = min; min = max; max = tmp; }

    return get_mask((max - min) + 1) << min;
}

inline uint64_t get_field(uint64_t value, unsigned min, unsigned max)
{
    if(min > max) { unsigned tmp = min; min = max; max = tmp; }

    return (value >> min) & get_mask((max - min) + 1);
}

inline uint64_t get_field(uint64_t value, unsigned bit)
{
    return get_field(value, bit, bit);
}

inline void set_field(uint64_t &value, unsigned min, unsigned max, uint64_t field)
{
    if(min > max) { unsigned tmp = min; min = max; max = tmp; }

    uint64_t mask = get_mask(min, max);
    value &= ~mask; value |= (field << min) & mask;
}

inline void set_field(uint64_t &value, unsigned bit, uint64_t field)
{
    set_field(value, bit, bit, field);
}

inline uint64_t get_bits(uint64_t value, unsigned min, unsigned max)
{
    return get_field(value, min, max);
}

inline uint64_t get_bits(uint64_t value, unsigned bit)
{
    return get_field(value, bit);
}

inline uint64_t get_some_bits(uint64_t value, unsigned start, unsigned size)
{
    start %= 64;
    if(start + size > 63) size = 63 - start;
    return get_field(value, start, start + size);
}

inline void set_bits(uint64_t &value, unsigned min, unsigned max, uint64_t field)
{
    set_field(value, min, max, field);
}

inline void set_bits(uint64_t &value, unsigned bit, uint64_t field)
{
    set_field(value, bit, field);
}

class Message;


/**
 * This template class represents a message field. The field is stored as an
 * array of 64-bit unsigned integers. The type parameter is used only for
 * printing fields. It may be used in methods like <code>get</code> and others,
 * but it is not implemented.
 */
template<typename T = uint64_t, size_t field_width = 64, uint64_t field_mask = 0xffffFFFFffffFFFFull, size_t array_size = 1>
class Field
{
public:
    friend class Message;

    Field(uint64_t value = 0):
    _short_name("<unknown>"),
    _base_name("<unknown>"),
    _full_name("<unknown>"),
    _compare(true),
    _randomize(true),
    _valid(array_size),
    _data(array_size),
    _width(field_width)
    {
        for(size_t i = 0; i < array_size; i++)
        {
            _data[i]  = value;
            _valid[i] = false;
        }
    }

    Field(const Field &rhs) { _copy(rhs); }

    virtual ~Field() {}

    operator uint64_t () const { return get();  }

    Field& operator  = (const Field &rhs)
        { for(size_t i = 0; i < array_size; i++) { set(rhs.get(i), i); }          return *this; }
    Field& operator |= (const Field &rhs)
        { for(size_t i = 0; i < array_size; i++) { set(get(i) | rhs.get(i), i); } return *this; }
    Field& operator &= (const Field &rhs)
        { for(size_t i = 0; i < array_size; i++) { set(get(i) & rhs.get(i), i); } return *this; }
    Field& operator ^= (const Field &rhs)
        { for(size_t i = 0; i < array_size; i++) { set(get(i) ^ rhs.get(i), i); } return *this; }

    /// Returns the size of the field's array.
    size_t size() const { return _data.size(); }

    /// Returns the field's data.
    const uint64_t* data() const { return _data.data(); } 

    unsigned get_width() const { return _width; }
    
    uint64_t& operator[] (size_t i)
    {
        CPPTESK_ASSERTION(i < _data.size(), "Index is out of bounds: "
            << "field="   << getBaseName()
            << ", size="  << std::dec << _data.size()
            << ", index=" << std::dec << i);

        _valid[i] = true;
        return _data[i];
    }

    uint64_t operator[] (size_t i) const
    {
        CPPTESK_ASSERTION(i < _data.size(), "Index is out of bounds: "
            << "field="   << getBaseName()
            << ", size="  << std::dec << _data.size()
            << ", index=" << std::dec << i);

        return _data[i];
    }

    uint64_t get(size_t i = 0) const
    {
        return this->operator[](i);
    }

    void set(uint64_t value, size_t i = 0)
    {
        CPPTESK_ASSERTION((value & ~get_mask(field_width)) == 0, "Cannot assign " << std::hex << value
            << " to " << getFullName() << "[" << std::dec << field_width << std::hex << "]");

        this->operator[](i) = (value & get_mask(field_width) & field_mask);

        _valid[i] = true;
    }

    uint64_t getBits(unsigned lo, unsigned hi) const { return get_field(_data[0], lo, hi); }
    uint64_t getBit(unsigned bit) const { return get_field(_data[bit/_width], bit - (bit/_width)*_width); }

    void setBits(unsigned lo, unsigned hi, uint64_t val) { set_field(_data[0], lo, hi, val); }
    void setBit(unsigned bit, uint64_t val) { set_field(_data[0], bit, val); }

    std::string getShortName() const { return _short_name; }
    std::string getBaseName()  const { return _base_name; }
    std::string getFullName()  const { return _full_name; }

    void parseFullName(const std::string &full_name)
    {
        size_t i;

        _set_full_name(full_name);

        if((i = full_name.find("::")) == std::string::npos)
        {
            _set_base_name("");
            _set_short_name(full_name);
        }
        else
        {
            _set_base_name(full_name.substr(0, i));
            _set_short_name(full_name.substr(i + 2));
        }
    }

    bool isComparable() const { return _compare; }
    void setComparable(bool compare = true) { _compare = compare; }

    bool isRandomizable() const { return _randomize; }
    void setRandomizable(bool randomize = true) { _randomize = randomize; }

    bool isValid() const
    {
        for(size_t i = 0; i < _data.size(); i++)
        {
            if(!_valid[i])
                { return false; }
        }
        return true;
    }

    void setValid(bool valid = true)
    {
        for(size_t i = 0; i < _data.size(); i++)
            { _valid[i] = valid; }
    }

    void invalidate() { setValid(false); }

    virtual std::string compare(const Field &rhs) const
    {
        if(!(isValid() && rhs.isValid()))
        {
            CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_WARN, getBaseName() << (!isValid() ? "{spec=" : "{impl=")
                << toString() << "} is uninitialized");

            return "";
        }

        return forceCompare(rhs);
    }

    virtual std::string forceCompare(const Field &rhs) const
    {
        std::stringstream buffer;

        // NOTICE: When calling compare() all the fields are always compared.
        // For incomparable fields compare() is not called by the message.
        for(size_t i = 0; i < size(); i++)
        {
            if(get(i) != rhs.get(i))
            {
                buffer << getBaseName();
                buffer  << "{spec=" << toString() << ", " << "impl=" << rhs.toString() << "}";
                break;
            }
        }

        return buffer.str();
    }

    /// Returns true if fields are equal (in terms of forceCompare()).
    bool equals(const Field &rhs) const
    {
        return forceCompare(rhs).empty();
    }

    virtual std::string toString() const
    {
        std::stringstream buffer;

        buffer << std::showbase << std::hex;
        buffer << (isComparable() ? "" : "[") << getShortName() << (isComparable() ? "" : "]") << "(";
        for(size_t i = 0; i < size(); i++)
        {
            if(i != 0)
                { buffer << ":"; }

            // Type casting is used to print fields properly.
            if(_valid[i])
                { buffer << (T)get(i); }
            else
                { buffer << "<unknown>"; }
        }
        buffer << ")";

        return buffer.str();
    }

    virtual std::string forceToString() const
    {
        std::stringstream buffer;

        buffer << std::showbase << std::hex;
        buffer << (isComparable() ? "" : "[") << getShortName() << (isComparable() ? "" : "]") << "(";
        for(size_t i = 0; i < size(); i++)
        {
            if(i != 0)
                { buffer << ":"; }

            // Type casting is used to print fields properly.
            buffer << (T)get(i);
        }
        buffer << ")";

        return buffer.str();
    }

    /// Randomizes the field data.
    virtual void randomize()
    {
        for(size_t i = 0; i < size(); i++)
            { set(CPPTESK_RANDOM_WIDTH(uint64_t, field_width), i); }
    }

private:
    void _set_short_name(const std::string &short_name)
    {
        _short_name = short_name;
    }

    void _set_base_name(const std::string &base_name)
    {
        _base_name  = base_name;
    }

    void _set_full_name(const std::string &full_name)
    {
        _full_name  = full_name;
    }

    void _copy(const Field &rhs)
    {
        _short_name = rhs._short_name;
        _base_name  = rhs._base_name;
        _full_name  = rhs._full_name;
        _compare    = rhs._compare;
        _randomize  = rhs._randomize;
        _valid      = rhs._valid;
        _data       = rhs._data;
        _width      = rhs._width;
    }

private:
    /// Field's short name.
    std::string _short_name;
    /// Message's name
    std::string _base_name;
    /// Field's full name.
    std::string _full_name;
    /// Compare or not.
    bool _compare;
    /// Randomize or not.
    bool _randomize;

    // IMPORTANT: Template-parameter-dependent fields should be declared in the
    // end of the template class. Otherwise, a compiler produces incorrect code

    /// Validity.
    std::vector<bool> _valid;
    /// Field's value.
    std::vector<uint64_t> _data;
    /// Field's width.
    int _width;
};

}} // namespace cpptesk::hw

