/*
 * Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: History of events (stimuli & reactions) 
///////////////////////////////////////////////////////////////////////////////

#include <hw/history.hpp>
#include <hw/unused.hpp>
#include <hw/sim_info.hpp>
#include <hw/diagnostics/diagnostics.hpp>
#include <hw/diagnostics/metrics/metric.hpp>
#include <hw/diagnostics/metrics/field_metric.hpp>
#include <hw/diagnostics/metrics/information_metric.hpp>
#include <hw/report/model.hpp>
#include <hw/report/console_printer.hpp>
#include <hw/report/xml_printer.hpp>
#include <hw/report/reaction_printer.hpp>

#include <iomanip>
#include <sstream>

#include <string.h>

namespace cpptesk {
namespace hw {

EventHistory::EventHistory()
{
    reset();
}

EventHistory::~EventHistory()
{
    reset();
}

event_ptr_t EventHistory::register_stimulus(const Stimulus &stimulus)
{
    event_ptr_t stimulus_ptr(new Stimulus(stimulus));

    stimulus_ptr->id(_id++);
    _stimuli.push_back(stimulus_ptr);
    _stimulus_count++;

    return stimulus_ptr;
}

event_ptr_t EventHistory::register_reaction(const Reaction &reaction)
{
    event_ptr_t reaction_ptr(new Reaction(reaction));

    reaction_ptr->id(_id++);
    _reactions.push_back(reaction_ptr);
    
    if(reaction.getError() != MISSING_REACTION)
    {
        _reaction_count[reaction.getError()]++;
    }
    else if(reaction.getSpecMessage()->isOptional())
    {
        _optional_missing++;
    }
    else
        if(reaction.isExpected())
            _expected_missing++;
        else
            _reaction_count[reaction.getError()]++;

    return reaction_ptr;
}

void EventHistory::reset_current_set()
{
    // Append the current set of reactions into the history.
    // Note that the history does not contain stimuli.
    _history.insert(_history.end(), _reactions.begin(), _reactions.end());

    // Reduce the history.
    while(_history.size() > HISTORY_SIZE)
        { _history.erase(_history.begin()); }

    // Clear the current set.
    _stimuli.clear();
    _reactions.clear();

}

bool EventHistory::verdict()
{
    bool result = true;

    // If all reactions are normal, then the verdict is true.
    for(reaction_list_t::iterator i = _reactions.begin(); i != _reactions.end(); i++)
    {
        const Reaction *reaction = dynamic_cast<const Reaction*>(i->get());

        // Otherwise, the verdict is false.
        if(reaction->getError() != NORMAL_REACTION)
        {
            result = false;
            aspect_error_traceFailure("error", reaction->getShortDescription().c_str(), reaction->getFullDescription().c_str());
        }
    }

    // The reactions have been analyzed and there is no reason to analyze them in the future.
    reset_current_set();

    return result;
}

std::string EventHistory::error(error_mode_t mode) const
{
    std::stringstream buffer;

    bool written  = false;
    bool finished = false;
    bool last_error_written      = false;
    bool last_real_error_written = false;

    // Analyze the current set of reactions.
    for(reaction_list_t::const_reverse_iterator i = _reactions.rbegin(); i != _reactions.rend(); i++)
    {
        const Reaction *reaction = dynamic_cast<const Reaction*>(i->get());

        if((reaction->isRealError() && (mode & LAST_ERROR || mode & ALL_ERRORS)) ||
           (reaction->isError() && reaction->isExpected() && (mode & LAST_EXPECTED || mode & ALL_EXPECTED)))
        {
            if(reaction->isRealError() && mode & LAST_ERROR && !(mode & ALL_ERRORS) && last_real_error_written)
                { continue; }
            if(reaction->isError() && reaction->isExpected() && mode & LAST_EXPECTED && !(mode & ALL_EXPECTED) && last_error_written)
                { continue; }

            if(written) { buffer << std::endl; }
            buffer << reaction->getFullDescription();

            if(reaction->isRealError() && mode & LAST_ERROR)
            {
                if((mode & ALL_EXPECTED) | (mode & LAST_EXPECTED))
                    { last_real_error_written = true; }
                else if(!(mode & ALL_ERRORS))
                    { finished = true; break; }
            }

            if(reaction->isError() && reaction->isExpected() && mode & LAST_EXPECTED)
            {
                if((mode & ALL_ERRORS) | (mode & LAST_ERROR))
                    { last_error_written = true; }
                else if(!(mode & ALL_EXPECTED))
                    { finished = true; break; }
            }

            if((mode == (error_mode_t)(LAST_EXPECTED | LAST_ERROR) && last_real_error_written && last_error_written))
                { finished = true; break; }

            written = true;
        }
    }

    if(finished)
        { return buffer.str(); }

    // Analyze the history of reactions
    for(reaction_list_t::const_reverse_iterator i = _history.rbegin(); i != _history.rend(); i++)
    {
        const Reaction *reaction = dynamic_cast<const Reaction*>(i->get());

        if((reaction->isRealError() && (mode & LAST_ERROR || mode & ALL_ERRORS)) ||
           (reaction->isError() && reaction->isExpected() && (mode & LAST_EXPECTED || mode & ALL_EXPECTED)))
        {
            if(reaction->isRealError() && mode & LAST_ERROR && !(mode & ALL_ERRORS) && last_real_error_written)
                { continue; }
            if(reaction->isError() && reaction->isExpected() && mode & LAST_EXPECTED && !(mode & ALL_EXPECTED) && last_error_written)
                { continue; }

            if(written) { buffer << std::endl; }
            buffer << reaction->getFullDescription();

            if(reaction->isRealError() && mode & LAST_ERROR)
            {
                if((mode & ALL_EXPECTED) | (mode & LAST_EXPECTED))
                    { last_real_error_written = true; }
                else if(!(mode & ALL_ERRORS))
                    { break; }
            }

            if(reaction->isError() && reaction->isExpected() && mode & LAST_EXPECTED)
            {
                if((mode & ALL_ERRORS) | (mode & LAST_ERROR))
                    { last_error_written = true; }
                else if(!(mode & ALL_EXPECTED))
                    { break; }
            }

            if((mode == (error_mode_t)(LAST_EXPECTED | LAST_ERROR) && last_real_error_written && last_error_written))
                { break; }

            written = true;
        }
    }

    return buffer.str();
}

unsigned EventHistory::count(error_type_t error) const
{
    return _reaction_count[error];
}

void EventHistory::reset()
{
    _id = 0;
    // Reset the current set of reactions.
    _reactions.clear();
    // Reset the history of reactions.
    _history.clear();
    // Reset statistics on stimuli.
    _stimulus_count = 0;
    // Reset statistics on reactions.
    ::memset(_reaction_count, 0, sizeof(_reaction_count));
    // Reset missing statistics.
    _optional_missing = 0;
    _expected_missing = 0;
}

void EventHistory::trace_statistics()
{
    unsigned max_width = 0;
    unsigned all_count = 0;

    std::stringstream buffer;

    for(unsigned error = 0; error <= __ERROR_TYPE_NUMBER; error == 0 ? error = 1 : error <<= 1)
    {
        const std::string &error_string = getErrorString((error_type_t)error);

        if(error_string.length() > max_width)
            { max_width = error_string.length(); }

        all_count += count((error_type_t)error);
    }

    buffer.width(max_width); buffer.fill(' ');
    buffer << std::left << "[STIMULUS]" << ": ";
    buffer << std::dec << _stimulus_count << std::hex << " (" << std::fixed << std::setprecision(2)
        << (all_count > 0 ? (float)_stimulus_count/(float)all_count : 0.0) << " S/R)" << std::endl;

    buffer.width(max_width);
    buffer << std::left << "[REACTION]" << ": ";
    buffer << std::dec << all_count << std::hex << " (" << std::fixed << std::setprecision(2)
        << (_stimulus_count > 0 ? (float)all_count/(float)_stimulus_count : 0.0) << " R/S)";

    for(unsigned error = 0; error <= __ERROR_TYPE_NUMBER; error == 0 ? error = 1 : error <<= 1)
    {
        buffer.width(max_width);
        buffer << std::endl;
        buffer << std::left << getErrorString((error_type_t)error) << ": ";
        buffer << std::dec << count((error_type_t)error) << std::hex;
    }

    CPPTESK_PRINT_SEPARATOR();
    CPPTESK_DEBUG_STREAM() << buffer.str() << std::endl;
}

void EventHistory::trace_diagnosis(uint64_t cycles, time_t time)//, const char* sub_name)
{
    // Analyze the history only.
    SimulationInfo sim_info;

    sim_info.cycles = cycles;
    sim_info.time = time;
    sim_info.reaction_timeout = _reaction_timeout;
    sim_info.unexpected_reaction_timeout = _unexpected_reaction_timeout;
    sim_info.stimuli = _stimulus_count;
    
    unsigned all_count = 0;
    
    for(unsigned error = 0; error < __ERROR_TYPE_NUMBER;  error == 0 ? error = 1 : error <<= 1)
    {
        all_count += count((error_type_t)error);

        switch(error)
        {
            case NORMAL_REACTION: 
                sim_info.reaction_normal = count((error_type_t)error);
                break;
            case MISSING_REACTION: 
                sim_info.reaction_missing = count((error_type_t)error);
                break;
            case INCORRECT_REACTION: 
                sim_info.reaction_incorrect = count((error_type_t)error);
                break;
            case UNEXPECTED_REACTION: 
                sim_info.reaction_unexpected = count((error_type_t)error);
                break;
            default:
                break;
        }
    }

    sim_info.optional_missing = _optional_missing;
    sim_info.expected_missing = _expected_missing;
    
    sim_info.reactions = all_count + _optional_missing + _expected_missing;

    cpptesk::hw::diagnostics::Diagnostics diagnostics(_history, sim_info);
    
    // Set metric for diagnostics
    cpptesk::hw::diagnostics::BaseMetric::metric = cpptesk::hw::diagnostics::FieldMetric::getMetric();

    cpptesk::hw::diagnostics::ReportUnit report_unit = diagnostics.run();
    
    cpptesk::hw::report::ConsolePrinter console_printer;
    console_printer.print(CPPTESK_DEBUG_STREAM(), report_unit.report, Debug::get().isColoredPrintEnabled());
    
    // Print all reactions in log file
    cpptesk::hw::report::ReactionPrinter reaction_printer = cpptesk::hw::report::ReactionPrinter();
    reaction_printer.print((char*)"diagnostic_trace.txt", report_unit.reaction_vector);

    // Print results in xml file
    cpptesk::hw::report::XMLPrinter xml_printer = cpptesk::hw::report::XMLPrinter();
    xml_printer.print("diagnostic_trace.xml", report_unit.report);
}

void EventHistory::setReactionTimeout(cycle_t timeout)
{
    _reaction_timeout = timeout;
}

void EventHistory::setUnexpectedReactionTimeout(cycle_t timeout)
{
    _unexpected_reaction_timeout = timeout;
}

}} // namespace cpptesk::hw
