/*
 * Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of a hardware interface
///////////////////////////////////////////////////////////////////////////////

#include <hw/assertion.hpp>
#include <hw/debug.hpp>
#include <hw/interface.hpp>
#include <hw/message.hpp>
#include <hw/process/process.hpp>

#include <algorithm>
#include <ostream>
#include <vector>

#define __CPPTESK_CHECK_ARBITER() \
   do { \
        if(!_arbiter) { \
            CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_WARN, "Reaction arbiter is not set for the interface " << *this); \
        } \
   } while(false)

namespace cpptesk {
namespace hw {

void Interface::request(const Process &process, const Message &message)
{
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[IFACE:REQUEST](iface=" << *this << ")");

    Component::request(process);
    _message = (Message*)&message;

    // Increment number of stimuli
    incrementEventNumber();
}

void Interface::capture()
{
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[IFACE:CAPTURE](iface=" << *this << ")");

    Component::capture();
 }

void Interface::release()
{
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[IFACE:RELEASE](iface=" << *this << ")");

    Component::release();
    _message = 0;
}

void Interface::appendRequest(const Process &process, const Message &message)
{
    getArbiter()->appendRequest(process, message);
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:APPEND](iface=" << *this << "): " << process);

    _request_count++;
}

bool Interface::chooseRequest(const Process &process, const Message &message)
{
    bool choose = getArbiter()->chooseRequest(process, message);
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:CHOOSE](iface=" << *this << "): " << process << ": " << (choose ? "YES" : "NO"));

    return choose;
}

bool Interface::preChooseRequest(const Process &process, const Message &message)
{
    bool choose = getArbiter()->preChooseRequest(process, message);
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:PRECHOOSE](iface=" << *this << "): " << process << ": " << (choose ? "YES" : "NO"));

    return choose;
}

void Interface::removeRequest(const Process &process, const Message &message)
{
    CPPTESK_ASSERTION(_request_count > 0, "There are no requests to remove (probably, WAIT_REACTION() is missing)");
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:REMOVE](iface=" << *this << "): " << process);

    getArbiter()->removeRequest(process, message);

    _request_count--;
}

void Interface::removeRequest(const Process &process, const Message &message, cycle_t timemark)
{
    removeRequest(process, message);

    _remove_reaction_timemark = timemark;
}

bool Interface::can_adapter_start(const Process &process, const Message &message)
{
    bool choose = isModelBased() ? getArbiter()->chooseRequest(process, message) :
                                   getArbiter()->preChooseRequest(process, message);

    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:CAN_ADAPTER_START](iface=" << *this << "): " <<
                                    process << ": " << (choose ? "YES" : "NO"));
    return choose;
}

bool Interface::takeReaction(cycle_t timemark)
{
    CPPTESK_ASSERTION(_request_count > 0, "There are no requests to take");

    // Increment number of reactions
    incrementEventNumber();

    // If a request has been taken, returns false
    if(_take_reaction_timemark == timemark)
        { return false; }

    _take_reaction_timemark = timemark;

    // Otherwise returns true
    return true;
}

void Interface::takeOldReaction()
{
    // Increment number of reactions
    incrementEventNumber();
}

bool Interface::find_impl_reaction(Message &message, Reaction &reaction)
{
    for(reactions_t::iterator i = _impl_reactions.begin(); i != _impl_reactions.end(); i++)
    {
        reaction = *i;
        if(message.compare(*(reaction.what())) == Message::COMPARE_OK)
        {
            CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[FIR] Possibly unexpected impl reaction " << message <<
                                            " seems to be really unexpected as it has not been removed by any spec"); 
            return true;
        }
        else
        {
            CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[FIR] Reactions are different in data: kept impl one " << *(reaction.what()) <<
                                            " is NOT equal to checked impl one " << message << "; both on " << *this << " iface");
        }
    }
    return false;
}

bool Interface::some_unexpected_reaction_matches(const cycle_t start_time, const cycle_t reaction_timeout,
                                                 const cycle_t unexpected_reaction_timeout, Message &message)
{
    // We can validate fields as we suggest this message to be fully correct - spec-msg,
    // but we receives it from deserializer having been invalidated
    message.validateFields();

    /* Check the unexpected reaction storage where a listener puts found unexpected reactions
       which have not violated impl reaction timeout */
    if(!_impl_reactions.empty())
    {
        for(reactions_t::iterator i = _impl_reactions.begin(); i != _impl_reactions.end(); i++)
        {
            Reaction &r = *i;
            /* t_sr - first point of specification message, delta_sr - timeout of specification message
               t_ir - first point of implementation message, delta_it - timeout of implementation message */
            cycle_t t_sr = start_time, delta_sr = reaction_timeout,
                    t_ir = r.when(),   delta_ir = unexpected_reaction_timeout;

            /* If the current model reaction is the same as some impl reaction having been put into the storage earlier */
            if((message.compare(*(r.what())) == Message::COMPARE_OK /*||
               (r.what())->compare(message) == Message::COMPARE_OK*/) /*&&
                 (chooseRequest(_process, *(r.what())))*/) // By now, we can't use this as FIFO Arbiter checks only pid
            {
               /* There should be an intersection between spec and impl reaction lifetime */
                if(t_ir <= t_sr + delta_sr && t_sr <= t_ir + delta_ir)
                {

                    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "[PURM] Possibly missing spec reaction " <<
                                        message << " is equal to impl reaction " << *(r.what()) << "; both on " << *this << " iface");

                    _impl_reactions.erase(i);
                    return true;
                }
                else
                {
                    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[PURM] Being equal in data, reactions are not good in timeouts: spec one " <<
                                        message << " is equal to impl one " << *(r.what()) << "; both on " << *this << " iface;" <<
                                        " timeouts: spec=" << delta_sr << ", impl=" << delta_ir << "; gentime: spec=" <<
                                        t_sr << ", impl=" << t_ir);
                }
            }
            else
            {
                CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[PURM] Reactions are different in data: spec one " << message <<
                                                " is NOT equal to impl one " << *(r.what()) << "; both on " << *this << " iface");
            }
        }
    }
    else
        CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[PURM] Possible unexpected reaction matcher didn't find any suitable reaction for " <<
                                        message << " because of the correspondent reaction list being empty on " << *this << " iface");

    // return to the initial state
    message.invalidateFields();

    return false;
}

bool Interface::some_spec_reaction_matches(Message &message)
{
    message.validateFields();
    if(!_spec_reactions.empty()) {
        reactions_t::iterator i;
	int first_optional = -1, first_non_optional = -1;
	int number;

        for(i = _spec_reactions.begin(), number = 0; i != _spec_reactions.end(); i++, number++) {
            Reaction &r = *i;
            Message &msg = const_cast<Message&>(*(r.what()));
            if(msg.compare(message) == Message::COMPARE_OK) { // use compare everywhere instead of forceCompare due to
                if(msg.isOptional())                          // usage of user-defined compare-functions
                {
                    if(first_optional == -1)
                        first_optional = number;
                }
                else
                {
                    if(first_non_optional == -1)
                        first_non_optional = number;
                }
            }
        }

        if(first_non_optional != -1)
            first_optional = first_non_optional;

        if(first_optional == -1)
	{
            CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[PMRM] Can't find appropriate spec reaction for impl reaction " << message << " on " << *this << " iface");
	    return false;
	}
        
        for(i = _spec_reactions.begin(), number = 0; number < first_optional; i++, number++) {}
        Reaction &r = *i;
	Message &msg = const_cast<Message&>(*(r.what()));

        CPPTESK_DEBUG_PRINT(DEBUG_INFO, "[PMRM] Possibly unexpected impl reaction " << message.forceToString() <<
                                        " is equal to spec reaction " << msg << "; both on iface " << *this <<
                                        "; remains: " << _spec_reactions.size());

        CPPTESK_DEBUG_PRINT(DEBUG_INFO, "[PMRM] Remove spec reaction " << r << "(" << msg << ", isOpt?" << (msg.isOptional() ? "yes" : "no") << ")" << "; size=" << _spec_reactions.size());

        _spec_reactions.erase(i);
        return true;
    }

    return false;
}

bool Interface::remove_spec_reaction_after_timeout(Message &message)
{
    message.validateFields();
    if(!_spec_reactions.empty()) {
        for(reactions_t::iterator i = _spec_reactions.begin(); i != _spec_reactions.end(); i++) {
            Reaction &r = *i;
            Message &msg = const_cast<Message&>(*(r.what()));
            if(msg.getId() == message.getId())
            {
                CPPTESK_DEBUG_PRINT(DEBUG_INFO, "[PMRM2] Correspondent spec reaction is found for " << message.forceToString() <<
                                                " (in spec_buffer: " << msg << ", id=" << msg.getId() << "); on iface " << *this <<
                                                "; remains: " << _spec_reactions.size());
                _spec_reactions.erase(i);
                return true;
	    }
        }
    }
    return false;
}


/* Put the message into spec-reaction-buffer to let impl-unexp-reactions be compared with */ \
void Interface::add_reaction_to_spec_reactions(Message &message, cycle_t time)
{
    message.setId(_current_id++);
    message.validateFields();
    Reaction spec_reaction = Reaction(MISSING_REACTION, *this, message, time, time, "");
    _spec_reactions.push_back(spec_reaction);
    message.invalidateFields();
    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, "Add specification reaction " << message.forceToString() << "(id=" << message.getId() << ")" <<
                                                 " to " << *this << ".spec_reactions: " << spec_reaction << "(is_opt?" << (message.isOptional() ? "yes":"no") << ")");
}

void Interface::add_reaction_to_impl_reactions(Message &message, cycle_t time)
{
    Reaction impl_reaction = Reaction(UNEXPECTED_REACTION, *this, message, time, time, "");
    _impl_reactions.push_back(impl_reaction);
    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, "Add possibly unexpected impl reaction " << message.forceToString() <<
                                                 " to " << *this << ".impl_reactions: " << impl_reaction);
}

}} // namespace cpptesk::hw
