/*
 * Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of an operation execution process
///////////////////////////////////////////////////////////////////////////////

#include <hw/assertion.hpp>
#include <hw/interface.hpp>
#include <hw/message.hpp>
#include <hw/process/process.hpp>
#include <hw/process/process_table.hpp>

#ifdef CPPTESK_CLONE_INPUT_MESSAGE
    #define _CPPTESK_CLONE_MESSAGE(message)  ((message)->clone())
    #define _CPPTESK_DELETE_MESSAGE(message) if(message) { delete (message); }
#else
    #define _CPPTESK_CLONE_MESSAGE(message)  (message)
    #define _CPPTESK_DELETE_MESSAGE(message)
#endif

namespace cpptesk {
namespace hw {

Process::Process(pid_t pid, ProcessTable &process_table):
    timeout(false),
    input_iface(0),
    input_message(0),
    original_message(0),
    listener(false),
    _pid(pid),
    _status(EXECUTING),
    _start_time(CPPTESK_INFINITY_TIME),
    _current_time(CPPTESK_INFINITY_TIME),
    _schedule_time(CPPTESK_INFINITY_TIME),
    _context(0),
    _attrs(),
    _process_table(process_table)
{
}

Process::Process(pid_t pid, cycle_t now, Interface &iface, Message &message, Attributes attrs, ProcessTable &process_table):
    timeout(false),
    input_iface(&iface),
    input_message(&message),
    original_message(&message),
    listener(false),
    _pid(pid),
    _status(EXECUTING),
    _start_time(now),
    _current_time(CPPTESK_INFINITY_TIME),
    _schedule_time(CPPTESK_INFINITY_TIME),
    _context(0),
    _attrs(attrs),
    _process_table(process_table)
{
    if(attrs.clone_message)
    {
        input_message = message.clone();
        CPPTESK_ASSERTION(input_message != 0, "Input message has not been cloned");
    }
}

Process::~Process()
{
    // If the message was cloned, it should deleted.
    if(input_message && _attrs.clone_message)
        { delete input_message; }
}

void Process::set_priority(unsigned v)
{
    // Change priority.
    _attrs.priority = v;
    // Inform the process table that the process priority has changed.
    _process_table.on_priority_change(*this);
}

void Process::resume()
    { _process_table.resume_process(*this); }
void Process::suspend()
    { _process_table.suspend_process(*this); }
void Process::suspend(cycle_t next_time_slot)
    { _process_table.suspend_process(*this, next_time_slot); }
void Process::schedule_process()
    { _process_table.schedule_process(*this); }
void Process::schedule_process(cycle_t time_slot)
    { _process_table.schedule_process(*this, time_slot); }

///////////////////////////////////////////////////////////////////////////////
// Cause-effect relation
///////////////////////////////////////////////////////////////////////////////

void Process::set_cause(const event_ptr_t &event)
{
    CPPTESK_ASSERTION(_cause.isNull(), "The cause of the process has been already set: " << *this);

    _cause = event;

    if(!_effects.empty())
        { (*_effects.begin())->cause(_cause); }
}

void Process::add_effect(event_ptr_t &event)
{
   if(_effects.empty())
   {
       if(!_cause.isNull())
           { event->cause(_cause); }
   }
   else
       { event->cause(*_effects.rbegin()); }

   _effects.push_back(event);
}

void Process::caused_by(const Process &process)
{
    if(!process._effects.empty())
        { set_cause(*process._effects.rbegin()); }
    else if(!process._cause.isNull())
        { set_cause(process._cause); }
}

}} // namespace cpptesk::hw

