/*
 * Copyright 2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Process scheduler
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/process/process.hpp>

#include <map>
#include <set>
#include <vector>

namespace cpptesk {
namespace hw {

/**
 * Process scheduler basic class.
 */
class ProcessScheduler
{
public:
    ProcessScheduler(): _clock(0) {}
    virtual ~ProcessScheduler() {}

    /// Schedules the process execution for the current time slot.
    void schedule_process(Process &process);
    /// Schedules the process execution for the given time slot.
    void schedule_process(Process &process, cycle_t time_slot);
    /// Freezes the process (stops the process execution until it is scheduled for a certain time slot).
    void stop_process(Process &process);

    /// Returns the current time.
    cycle_t time() const { return _clock; }
    /// Returns the earliest time slot.
    cycle_t next_time() const;
    /// Starts processing of the given time slot. Returns the earliest non-empty time slot.
    cycle_t start_cycle(cycle_t now);
    
    /// Returns true iff the current delta cycle is completed.
    virtual bool is_cycle_completed() const = 0;
    /// Returns the next process to be executed at the current cycle.
    virtual Process& next_process() = 0;
    /// Adds the process to the current-cycle set of processes.
    virtual void activate_process(Process &process) = 0;
    /// Removes the process from the current-cycle set of processes.
    virtual void deactivate_process(Process &process) = 0;
    /// Handles priority change.
    virtual void on_priority_change(Process &process) = 0;

protected:
    /// Should be called in activate_process().
    void start_process_cycle(Process &process) { process.current_time(_clock); }

    typedef std::map<pid_t, Process*>       _processes_t;
    typedef std::map<cycle_t, _processes_t> _schedule_t;

    cycle_t     _clock;
    _schedule_t _schedule;
};

/**
 * Default process scheduler.
 */
class DefaultProcessScheduler: public ProcessScheduler
{
public:
    DefaultProcessScheduler(): ProcessScheduler(), _modified(false) {}
    virtual ~DefaultProcessScheduler() {}

    /// Returns true iff the current delta cycle is completed.
    virtual bool is_cycle_completed() const;
    /// Returns the next process to be executed at the current cycle.
    virtual Process& next_process();
    /// Adds the process to the current-cycle set of processes.
    virtual void activate_process(Process &process);
    /// Removes the process from the current-cycle set of processes.
    virtual void deactivate_process(Process &process);
    /// Handles priority change.
    virtual void on_priority_change(Process &process);

private:
    // Vector is used for effective sorting of processes.
    typedef std::vector<Process*> _process_set_t;
    // Set is used to avoid multiple occurences of the same process.
    typedef std::set<Process*>    _check_set_t;

    bool _modified;
    _process_set_t _current_set;
    _check_set_t _check_set;
};

}} // namespace cpptesk::hw
