/*
 * Copyright 2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Process table
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/process/context_switcher.hpp>
#include <hw/process/process.hpp>
#include <hw/process/process_scheduler.hpp>
#include <hw/process/process_stack.hpp>
#include <hw/time.hpp>
#include <hw/utils/hash_table.hpp>

#include <list>
#include <map>

namespace cpptesk {
namespace hw {

class Process;
class Interface;
class Message;

/**
 * Process table.
 */
class ProcessTable
{
public:
    typedef ContextSwitcher::method_t method_t;

    enum exec_mode_t
    {
        EXECUTE_CYCLE,
        EXECUTE_QUANTUM
    };

    ProcessTable();
    ProcessTable(ProcessScheduler &process_scheduler, ContextSwitcher &context_switcher);
    ~ProcessTable();
    
    /// Creates a new process.
    Process& make_process(BaseModel &model, method_t method,
        Interface &iface, Message &message, Process::Attributes &attrs);

    /// Start the process.
    void start_immediately(Process &process)
        { schedule_process(process); resume_process(process); }
    void start_this_cycle(Process &process)
        { schedule_process(process); }
    void start_next_cycle(Process &process)
        { schedule_process(process, _clock + 1); }

    /// Kills the process (terminates the process execution and removes all associated data).
    void kill_process(Process &process);
    /// Kills all processes.
    void kill_all();

    /// Resumes the process execution (executes the next process quantum).
    void resume_process(Process &process);
    /// Suspends the process execution and leaves the process in the
    /// current time slot (should be explicitly called by the process itself).
    void suspend_process(Process &process);
    /// Suspends the process execution and schedules its execution for the
    /// given time slot (should be explicitly called by the process itself).
    void suspend_process(Process &process, cycle_t next_time_slot);

    /// Schedules the process execution for the current time slot.    
    void schedule_process(Process &process)
        { _process_scheduler.schedule_process(process); }
    /// Schedules the process execution for the given time slot.
    void schedule_process(Process &process, cycle_t time_slot)
        { _process_scheduler.schedule_process(process, time_slot); }
    /// Freezes the process (stops the process execution until it is scheduled for a certain time slot).
    void stop_process(Process &process)
        { _process_scheduler.stop_process(process); }

    /// Handles priority change.
    void on_priority_change(Process &process)
        { _process_scheduler.on_priority_change(process); }

    /// Returns the process table size (number of running processes).
    size_t size() const { return _process_table.size(); }

    /// Returns the current time (set by cycle()).
    cycle_t time() const { return _clock; }
    /// Depending on the mode either simulates a cycle (composes a set of active processes and executes
    /// their stages in some order) or simulates a single process quantum (performs one context switch).
    /// Returns the time of the earliest future event (to make it possible to skip useless cycles).
    cycle_t execute(cycle_t now, exec_mode_t mode);

    /// Returns true iff the process is alive (exists in the process table).
    bool is_process_alive(pid_t pid) { return _process_table.find(pid) != _process_table.end(); }
    /// Returns the process with the given identifier.
    Process& get_process(pid_t pid) { return **_process_table.find(pid); }

    /// Returns the reference to the process stack.
    process_stack_t& process_stack() { return _process_stack; }
    /// Returns the call descriptor.
    Process::CallDescriptor get_call() const { return _call; }
    /// Prepares the call descriptor.
    void prepare_call(const Process::CallDescriptor &call) { _call = call; }
    /// Resets the call descriptor.
    void reset_call() { _call.reset(); }

private:
    static const size_t HASH_TABLE_SIZE = (4*1024);

private:
    void _make_main_process();

private:
    cycle_t _clock;
    pid_t _pid;
    Process _main_process;
    bool _delete_process_scheduler;
    ProcessScheduler &_process_scheduler;
    bool _delete_context_switcher;
    ContextSwitcher &_context_switcher;
    HashTable<pid_t, Process*> _process_table;
    process_stack_t _process_stack;

    /**
     * Descriptor of a current process call:
     *
     * CALL_PROCESS():  prepare_call(DESCRIPTOR);
     * start():         process.prepare_call(get_call()); reset_call();
     * START_PROCESS(): process.start_call();
     * STOP_PROCESS():  process.stop_call();
     */
    Process::CallDescriptor _call;

    // Debugging counts
    unsigned _created;
    unsigned _deleted;
};

}} // namespace cpptesk::hw
