/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Trace printer for Diagnostics Reactions
///////////////////////////////////////////////////////////////////////////////

#include <hw/report/reaction_printer.hpp>

#include <fstream>

#include <hw/component.hpp>
#include <hw/event.hpp>

#include <hw/diagnostics/reaction_unit.hpp>

namespace cpptesk {
namespace hw {
namespace report {

ReactionPrinter::ReactionPrinter()
{
    _file_trace = new std::fstream();
}

ReactionPrinter::~ReactionPrinter()
{
    delete _file_trace;
}

void ReactionPrinter::print(std::string _file_name, const cpptesk::hw::diagnostics::ReactionUnitVector &reactions)
{
    _file_trace->open(_file_name.c_str(), std::fstream::out | std::fstream::trunc);
    print(*_file_trace, reactions);
    _file_trace->close();
}

void ReactionPrinter::print(std::ostream &out, const cpptesk::hw::diagnostics::ReactionUnitVector &reactions)
{
    _reaction_units = reactions;
    
    out << "Printing vector of reactions from diagnostic module:" << std::endl << std::endl;
    
    for(uint i = 0; i < _reaction_units.size(); i++)
    {
        cpptesk::hw::diagnostics::ReactionUnit &ru = _reaction_units[i];
        cpptesk::hw::Reaction &reaction = ru.getReaction();

        // Title fo reaction
        out << "Reaction id = " << i << (reaction.isExpected()?" [Is EXPECTED]":"") << ", (";
        out << "Impl is " << (reaction.getImplMessage()->isOptional()? "[Optional]":"[NOT_Optional]") << "; ";
        out << "Spec is " << (reaction.getSpecMessage()->isOptional()? "[Optional]":"[NOT_Optional]");
        out << ")" << std::endl;

        // Description of reaction
        _describe_reaction(out, i);

        if(ru.get_parent_spec_id() != -1 && ru.get_parent_impl_id() != -1)// && debug)
        {
            _describe_parent(out, ru.get_parent_spec_id(), true);
            _describe_parent(out, ru.get_parent_impl_id(), false);
        }
        
        out << std::endl;
    }
}

void ReactionPrinter::_describe_reaction(std::ostream &out, uint id)
{
    cpptesk::hw::diagnostics::ReactionUnit &ru = ReactionPrinter::_reaction_units[id];
    cpptesk::hw::Reaction &reaction = ru.getReaction();

    if(reaction.getError() == MISSING_REACTION)
    {
        out << reaction.getShortDescription() << std::endl;
        out << "    SpecMessage: " << reaction.getSpecMessage()->forceToString() << std::endl;
        
        _print_explanation(out, reaction);
    }
    else if (reaction.getError() == UNEXPECTED_REACTION)
    {
        out << reaction.getShortDescription() << std::endl;
        out << "    ImplMessage: " << reaction.getImplMessage()->forceToString() << std::endl;
    }
    else
    {
        out << reaction.getShortDescription() << std::endl;
        
        out << "    SpecMessage: " << reaction.getSpecMessage()->forceToString() << std::endl;
        _print_explanation(out, reaction);
        
        out << "    ImplMessage: " << reaction.getImplMessage()->forceToString() << std::endl;
    }

    out << "Collapsed: " << (ru.is_collapsed() ? "[YES]" : "[NO]") << ", " <<
              "Collapse Rule: " << ru.get_collapse_rule() << std::endl;
}

void ReactionPrinter::_describe_parent(std::ostream &out, uint id, bool first)
{
    cpptesk::hw::diagnostics::ReactionUnit &ru = ReactionPrinter::_reaction_units[id];
    cpptesk::hw::Reaction &reaction = ru.getReaction();
    std::stringstream stream;
    
    const ::cpptesk::hw::Interface *iface = reaction.where();
    
    out << "  Parent-" << (first ? "1" : "2") <<
              " (reaction id=" << id << ") on iface=" << ((Component*)iface)->getName();
    _describe_reaction(out, id);
}

// For spec reactions only
void ReactionPrinter::_print_explanation(std::ostream &out, Reaction &reaction)
{
    std::string explanation = reaction.getSpecMessage()->getExplanation();
    if(explanation.length() > 0)
        out << "Explanation are: " << explanation << std::endl;
}

}}}
