/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Temporal sequence delay 
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/utils/range.hpp>

namespace cpptesk {
namespace hw {
namespace sequence {

////////////////////////////////////////////////////////////////////////////////
// Delay 
////////////////////////////////////////////////////////////////////////////////

/**
 * Delay is a range of unsigned integer values.
 */
class Delay: public Range<unsigned>
{
public:
    Delay(): Range<unsigned>(1) {}
    Delay(unsigned delay): Range<unsigned>(delay) {}
    Delay(unsigned min, unsigned max): Range<unsigned>(min, max) {}
};

/// Infinite time.
extern unsigned inf;

/// Infinity time: infinity() = inf.
inline unsigned infinity()
{
    return Delay::INF;
}

/// Temporal operator delay(n) means an n-cycles delay.
/// Example: E1 << delay(2) << E2.
inline Delay delay(unsigned n)
{
    return Delay(n);
}

/// Temporal operator delay(m,n) means any delay from the range [m,n].
/// Example: E1 << delay(1, 3) << E2.
inline Delay delay(unsigned m, unsigned n)
{
    return Delay(m, n);
}

/// Temporal operator no_delay() means absense of delay.
/// Example: E1 << no_delay() << E2.
inline Delay no_delay()
{
    return Delay(0);
}

/// Temporal operator any_delay() is equivalent to delay(0, inf).
/// Example: E1 << any_delay() << E2.
inline Delay any_delay()
{
    return Delay(0, Delay::INF);
}

/// Temporal operator delay() is equivalent to delay(1, inf).
/// Example: E1 << delay() << E2.
inline Delay delay()
{
    return Delay(1, Delay::INF);
}

/// Temporal operator 'at_least(n)' is equivalent to delay(n, inf).
/// Example: E1 << at_least(2) << E2.
inline Delay at_least(unsigned n)
{
    return Delay(n, Delay::INF);
}

/// Temporal operator 'at_most(n)' is equivalent to delay(0, n).
/// Example: E1 << at_most(100) << E2.
inline Delay at_most(unsigned n)
{
    return Delay(0, n);
}

}}} // namespace cpptesk::hw::sequence

