/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Sequence monitor 
///////////////////////////////////////////////////////////////////////////////

#include <hw/sequence/monitor.hpp>

namespace cpptesk {
namespace hw {
namespace sequence {

SequenceMonitor::SequenceMonitor(const Sequence &sequence, type_t type):
    sequence(sequence), type(type), success_count(0), failure_count(0), first_cycle(true) {}

SequenceMonitor::~SequenceMonitor()
{
    for(recognizers_t::iterator i = recognizers.begin(); i != recognizers.end(); i++)
        { delete *i; }
}

unsigned SequenceMonitor::cycle()
{
    unsigned new_match = 0;

    if(type != ONE_TIME || first_cycle)
    {
        // TODO: Do not create new recognizer if the previously launched one is blocked on the first guard
        recognizers.push_back(new SequenceRecognizer(sequence));
    }

    for(recognizers_t::iterator i = recognizers.begin(); i != recognizers.end();)
    {
        SequenceRecognizer *recognizer = *i;
        CPPTESK_ASSERTION(recognizer, "Internal error: recognizer is null");

        // Call onSuccess() if a new match is found
        unsigned new_inner_match = recognizer->cycle();

        new_match     += new_inner_match;
        success_count += new_inner_match;

        while(new_inner_match--)
            { onSuccess(*recognizer); }

        // The recognizer has been completed
	if(recognizer->isFinished())
        {
	    // Call onFailure() if the recognizer is failed
            if(recognizer->isFailed())
            {
                failure_count++;
                onFailure(*recognizer);
            }

            // Remove the recognizer from the list
            i = recognizers.erase(i);
        }
	else
        {
            // Keep the recognizer in the list
            i++;
        }
    }

    first_cycle = false;

    return new_match;
}

size_t SequenceMonitor::size() const
{
    size_t count = 0;

    for(recognizers_t::const_iterator i = recognizers.begin(); i != recognizers.end(); i++)
        { count += (*i)->size(); }

    return count;
}

}}} // namespace cpptesk::hw::sequence

