/*-
 * Copyright 1997-2003 John-Mark Gurney.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <limits.h>
#include <stdlib.h>
#include <atl/pqueue.h>
#include <atl/arraylist.h>

/******************************************************************************/
/**                            PQueue Data Types                             **/
/******************************************************************************/

struct PQueue
{
    int size;
    int dlog;
    PQueueItem **cons;
    PQueueItem *min;
    PQueueItem *root;
    int heap_alloc_by;
    int use_free_items;
    ArrayList heap_items;
    ArrayList free_items;
};

struct PQueueItem
{
    int deg;
    int mark;
    PQueueItem *parent;
    PQueueItem *child;
    PQueueItem *left;
    PQueueItem *right;
    int key;
    void *data;
};

/******************************************************************************/
/**                        PQueue Internal Functions                         **/
/******************************************************************************/

static int ceillog2(unsigned int a)
{
    int o = a;
    int b = (8 * sizeof(int)) / 2;
    int i = 0;

    while(b)
    {
        i = (i << 1);
        if(a >= (1 << b))
        {
            a /= (1 << b);
            i = i | 1;
        }
        else
        {
            a &= (1 << b) - 1;
        }
        b /= 2;
    }
    if((1 << i) == o)
        { return i; }
    else
        { return i + 1; }
}

static void init_PQueueItem(PQueueItem *item, int key, void *data)
{
    item->deg    = 0;
    item->mark   = 0;
    item->parent = NULL;
    item->child  = NULL;
    item->left   = item;
    item->right  = item;
    item->key    = key;
    item->data   = data;
}

static PQueueItem* remove_PQueueItem(PQueueItem *item)
{
    PQueueItem *ret;

    if(item == item->left)
        { ret = NULL; }
    else
        { ret = item->left; }

    if(item->parent != NULL && item->parent->child == item)
        { item->parent->child = ret; }

    item->right->left = item->left;
    item->left->right = item->right;

    item->parent = NULL;
    item->left   = item;
    item->right  = item;

    return ret;
}

static void insertAfter_PQueueItem(PQueueItem *lhs, PQueueItem *rhs)
{
    if(lhs == lhs->right)
    {
        lhs->right = rhs;
        lhs->left  = rhs;
        rhs->right = lhs;
        rhs->left  = lhs;
    }
    else
    {
        rhs->right = lhs->right;
        lhs->right->left = rhs;
        lhs->right = rhs;
        rhs->left  = lhs;
    }
}

static void insertBefore_PQueueItem(PQueueItem *lhs, PQueueItem *rhs)
{
    insertAfter_PQueueItem(lhs->left, rhs);
}

static void linkChild_PQueueItem(PQueueItem *child, PQueueItem *parent)
{
    if(parent->child == NULL)
        { parent->child = child; }
    else
        { insertBefore_PQueueItem(parent->child, child); }

    child->parent = parent;
    parent->deg++;
    child->mark = 0;
}

static void insertRootList_PQueue(PQueue *pqueue, PQueueItem *item)
{
    if(pqueue->root == NULL)
    {
        pqueue->root = item;
        item->left   = item;
        item->right  = item;
    }
    else
    {
        insertAfter_PQueueItem(pqueue->root, item);
    }
}

static void removeRootList_PQueue(PQueue *pqueue, PQueueItem *item)
{
    if(item->left == item)
        { pqueue->root = NULL; }
    else
        { pqueue->root = remove_PQueueItem(item); }
}

static void checkCons_PQueue(PQueue *pqueue)
{
    int dlog;

    if(pqueue->dlog == -1 || pqueue->size > (1 << pqueue->dlog))
    {
        dlog = pqueue->dlog;

        if((pqueue->dlog = ceillog2(pqueue->size) + 1) < 8)
            { pqueue->dlog = 8; }

        if(dlog != pqueue->dlog)
        {
            pqueue->cons = (PQueueItem **)realloc(pqueue->cons, sizeof(PQueueItem*) * (pqueue->dlog + 1));
            assertion(pqueue->cons != NULL, "checkCons_PQueue: pqueue->cons == NULL");
        }
    }
}

static void consolidate_PQueue(PQueue *pqueue)
{
    int i, d, D;
    PQueueItem **a;
    PQueueItem *w;
    PQueueItem *y;
    PQueueItem *x;

    checkCons_PQueue(pqueue);

    D = pqueue->dlog + 1;
    a = pqueue->cons;

    for(i = 0; i < D; i++)
        { a[i] = NULL; }

    while((w = pqueue->root) != NULL)
    {
        x = w;
        removeRootList_PQueue(pqueue, w);
        d = x->deg;

        while(a[d] != NULL)
        {
            y = a[d];
            if(x->key > y->key)
                { PQueueItem *t = x; x = y; y = t; }
            linkChild_PQueueItem(y, x);
            a[d] = NULL;
            d++;
        }
        a[d] = x;
    }
    pqueue->min = NULL;
    for(i = 0; i < D; i++)
    {
        if(a[i] != NULL)
        {
            insertRootList_PQueue(pqueue, a[i]);

            if(pqueue->min == NULL || a[i]->key < pqueue->min->key)
                { pqueue->min = a[i]; }
        }
    }
}

static PQueueItem* extractMinItem_PQueue(PQueue *pqueue)
{
    PQueueItem *ret = pqueue->min;
    PQueueItem *x, *y, *orig = NULL;

    /* put all the children on the root list */
    /* for true consistancy, we should use remove_PQueueItem */
    for(x = ret->child; x != orig && x != NULL;)
    {
        if(orig == NULL)
            { orig = x; }
        y = x->right;
        x->parent = NULL;
        insertRootList_PQueue(pqueue, x);
        x = y;
    }
    
    removeRootList_PQueue(pqueue, ret);
    pqueue->size--;

    if(pqueue->size == 0)
        { pqueue->min = NULL; }
    else
    {
        pqueue->min = ret->right;
        consolidate_PQueue(pqueue);
    }

    return ret;
}

/******************************************************************************/
/**                        PQueue Interface Functions                        **/
/******************************************************************************/

PQueue *create_PQueue(int heap_alloc_by, int use_free_items)
{
    PQueue *pqueue = (PQueue*)malloc(sizeof(PQueue));
    assertion(pqueue != NULL, "create_PQueue: out of memory");

    pqueue->size = 0;
    pqueue->dlog = -1;
    pqueue->cons = NULL;
    pqueue->min  = NULL;
    pqueue->root = NULL;

    pqueue->heap_alloc_by  = heap_alloc_by;
    pqueue->use_free_items = use_free_items;

    init_ArrayList(&pqueue->heap_items, sizeof(PQueueItem),  heap_alloc_by, NULL);
    init_ArrayList(&pqueue->free_items, sizeof(PQueueItem*), heap_alloc_by, NULL);

    return pqueue;
}

void destroy_PQueue(PQueue *pqueue)
{
    if(pqueue->cons != NULL)
        { free(pqueue->cons); }

    pqueue->cons = NULL;

    destroy_ArrayList(&pqueue->heap_items);
    destroy_ArrayList(&pqueue->free_items);

    free(pqueue);
}

void insert_PQueue(PQueue *pqueue, int key, void *data)
{
    PQueueItem  *item;
    PQueueItem **item_holder;

    if(pqueue->free_items.size == 0)
    {
        item = (PQueueItem*)append_ArrayList(&pqueue->heap_items);
        assertion(item != NULL, "insert_PQueue: item == NULL");
    }
    else
    {
        item_holder = (PQueueItem**)get_ArrayList(&pqueue->free_items, pqueue->free_items.size - 1);
        remove_ArrayList(&pqueue->free_items, item_holder);

        item = *item_holder;
    }

    init_PQueueItem(item, key, data);
    insertRootList_PQueue(pqueue, item);

    if(pqueue->min == NULL || item->key < pqueue->min->key)
        { pqueue->min = item; }

    pqueue->size++;
}

void* min_PQueue(PQueue *pqueue)
{
    if(pqueue->min == NULL)
        { return NULL; }

    return pqueue->min->data;
}

void* extractMin_PQueue(PQueue *pqueue)
{
    PQueueItem  *item;
    PQueueItem **item_holder;

    if(pqueue->min != NULL)
    {
        item = extractMinItem_PQueue(pqueue);

        // Append the free block to the array list
        if(pqueue->use_free_items)
        {
            item_holder = (PQueueItem**)append_ArrayList(&pqueue->free_items);
           *item_holder = item;
        }
    }

    return item->data;
}

PQueue* union_PQueue(PQueue *lhs, PQueue *rhs)
{
    PQueueItem *item;

    if(lhs->root == NULL || rhs->root == NULL)
    {
        if(lhs->root == NULL)
            { destroy_PQueue(lhs); return rhs; }
        else
            { destroy_PQueue(rhs); return lhs; }
    }

    lhs->root->left->right = rhs->root;
    rhs->root->left->right = lhs->root;
    item = lhs->root->left;
    lhs->root->left = rhs->root->left;
    rhs->root->left = item;
    lhs->size += rhs->size;

    if(rhs->min->key < lhs->min->key)
        { lhs->min = rhs->min; }

    destroy_PQueue(rhs);
    return lhs;
}

