/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <rand/lagged_fibonacci.h>
#include <utils/assertion.h>

#include <stdlib.h>
#include <string.h>

///////////////////////////////////////////////////////////////////////////////
// Additive lagged Fibonacci random number generator
///////////////////////////////////////////////////////////////////////////////

void init_LFG(LFG *generator)
{
    assertion(generator != NULL, "init_LFG: generator == NULL");

    seed_LFG(generator, 0);

    generator->j = (RAND_LFG_K - RAND_LFG_J);
    generator->k = 0;
}

void seed_LFG(LFG *generator, uint32_t seed)
{
    int i, j, k;
    char b[(32 - 1) * (RAND_LFG_K - 1)];

    memset(generator->state, 0, sizeof(generator->state));

    // See, e.g., Pat Burns's "Lagged, Fibonacci Random Number Generators":
    // l.s.b. of the 11th word should be 1.
    generator->state[(RAND_LFG_K - 1) - 11] |= 1;

    seed ^= RAND_LFG_SEED;

    for(i = 0; i < 32; i++)
        { b[i] = (seed >> i) & 1; }

    // The initial state is filled with bits generated by the Tausworthe binary
    // shift register: b[n] = (b[n-31] + b[n-6] + b[n-4] + b[n-2] + b[n-1]) mod 2.
    for(i = 31; i < sizeof(b); i++)
        { b[i] = (b[i-31] + b[i-6] + b[i-4] + b[i-2] + b[i-1]) & 1; }

    for(i = 0; i < sizeof(b); i++)
    {
        j = (RAND_LFG_K - 1) - (i % (RAND_LFG_K - 1));
        k = i / (RAND_LFG_K - 1) + 1;

        if(b[i])
            { generator->state[j] |= (1 << k); }
    }
}

uint32_t next_LFG(LFG *generator)
{
    // x[n] = x[n-j] + x[n-k].
    const uint32_t next = generator->state[generator->j]
                        + generator->state[generator->k];

    // LFG state is organized as a cyclic array.
    generator->state[generator->k] = next;

    if(++generator->j == RAND_LFG_K /* array size */) { generator->j = 0; }
    if(++generator->k == RAND_LFG_K /* array size */) { generator->k = 0; }

    return next;
}

///////////////////////////////////////////////////////////////////////////////
// Modified additive lagged Fibonacci random number generator
///////////////////////////////////////////////////////////////////////////////

void init_MLFG(MLFG *generator)
{
    assertion(generator != NULL, "init_MLFG: generator == NULL");

    init_LFG(&generator->x);
    init_LFG(&generator->y);
    seed_LFG(&generator->y, 1);
}

void seed_MLFG(MLFG *generator, uint32_t seed)
{
    seed_LFG(&generator->x, seed);
    seed_LFG(&generator->y, seed + 1);
}

uint32_t next_MLFG(MLFG *generator)
{
    const uint32_t x = next_LFG(&generator->x);
    const uint32_t y = next_LFG(&generator->y);

    // Avoid bit correlations.
    return (x & ~1) ^ (y >> 1);
}

#ifdef TEST_RAND_LFG
int main()
{
    int i, j;
    int x[64];
    MLFG generator;

    for(i = 0; i < 1; i++)
    {
        memset(x, 0, sizeof(x));

        init_MLFG(&generator);
        seed_MLFG(&generator, i);

        printf("seed=%x\n", i);
        for(j = 0; j < 10000 * (sizeof(x)/sizeof(int)); j++)
        {
            const uint32_t next = next_MLFG(&generator);

            x[next % (sizeof(x)/sizeof(int))]++;
 
            printf("next=%x\n", next);
        }

        for(j = 0; j < sizeof(x)/sizeof(int); j++)
            { printf("%3d: %d\n", j, x[j]); }
    }

    return 0;
}
#endif // TEST_RAND_LFG

