/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <ctype.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <atl/arraylist.h>
#include <atl/integer.h>
#include <atl/int2ptrhash.h>
#include <atl/hashmap.h>
#include <atl/pqueue.h>
#include <sgl/graph.h>
#include <utils/assertion.h>

//------------------------------------------------------------------------------
// Graph Data Types
//------------------------------------------------------------------------------
#define SGL_NODES_ALLOC_BY     16536
#define SGL_NODES_HASH_SIZE    131072
#define SGL_EDGES_ALLOC_BY     32
#define SGL_END_NODES_ALLOC_BY 1

#define SGL_PF_MAX_HASH_SIZE   131072
#define SGL_PF_MIN_HASH_SIZE   1024
#define SGL_PF_PATH_ALLOC_BY   1024
#define SGL_PF_NULL_EDGE       ((SGL_Edge*)-1)

#define sgl_malloc             malloc
#define sgl_free               free

// Dump a state graph in the graphviz format (http://www.graphviz.org)
#define SGL_GRAPHVIZ_FORMAT

typedef struct SGL_PathFinder
{
    /// The source node.
    SGL_Node *source_node;
    /// The target node (it might be NULL).
    SGL_Node *target_node;
    /// The nodes having been explored.
    ArrayList nodes;
    /// Maps node indices into incoming edges.
    Int2PtrHashMap nodes_map;
    /// The priority queue directing the traversal order.
    PQueue *nodes_queue;
} SGL_PathFinder;

//------------------------------------------------------------------------------
// Graph Node Functions
//------------------------------------------------------------------------------

SGL_Node* sgl_alloc_Node(void)
{
    SGL_Node *node = (SGL_Node*)sgl_malloc(sizeof(SGL_Node));
    assertion(node != NULL, "sgl_alloc_Node: out of memory");

    return node;
}

void sgl_dealloc_Node(SGL_Node* node)
{
    assertion(node != NULL, "sgl_dealloc_Node: NULL pointer");
    sgl_free(node);
}

void sgl_init_Node(SGL_Node *node, int id, SGL_NodeId *node_id, SGL_Graph *graph)
{
    node->id      = id;
    node->node_id = r(node_id);

    node->weight   = SGL_INITIAL_WEIGHT;
    node->distance = 0;

    node->new_edge_count =  0;
    node->new_edge_index = -1;
    
    init_ArrayList(&node->edges, sizeof(SGL_Edge), SGL_EDGES_ALLOC_BY, (Destructor)sgl_destroy_Edge);

    node->graph = graph;
    node->data  = NULL;
}

void sgl_initWithData_Node(SGL_Node *node, int id, SGL_NodeId *node_id, SGL_Graph *graph, void *data)
{
    sgl_init_Node(node, id, node_id, graph);
    node->data = data;
}

void sgl_destroy_Node(SGL_Node *node)
{
    SGL_DataDestructor destructor = node->graph->node_destructor;

    destroy(node->node_id);
    destroy_ArrayList(&node->edges);

    if(destructor != NULL)
        { destructor(node->data); }
}

SGL_Edge* sgl_findEdge_Node(SGL_Node *node, SGL_StimulusId stimulus_id)
{
    if(sgl_getDegree_Node(node) <= stimulus_id.index)
        { return NULL; }

    return sgl_getEdge_Node(node, stimulus_id.index);
}

SGL_Edge* sgl_addEdge_Node(SGL_Node *start_node, SGL_StimulusId stimulus_id,
    SGL_Node *end_node, SGL_UpdateStatus *status)
{
    int i;
    SGL_Node **node_holder;
    SGL_Edge *edge = sgl_findEdge_Node(start_node, stimulus_id);

   *status = SGL_US_OLD_ALL;

    if(edge == NULL)
    {
        while(sgl_getDegree_Node(start_node) <= stimulus_id.index)
        {
            edge = (SGL_Edge*)append_ArrayList(&start_node->edges);
            sgl_init_Edge(edge, NULL, stimulus_id, NULL);
        }
    }

    if(!sgl_isValid_Edge(edge))
    {
        sgl_init_Edge(edge, start_node, stimulus_id, end_node);
        
        if(end_node == NULL)
        {
            if(start_node->new_edge_count++ == 0)
                { start_node->new_edge_index = stimulus_id.index; }
        }

       *status |= SGL_US_NEW_EDGE;
        return edge;    
    }

    if(end_node == NULL)
        { return edge; }

    if(sgl_isNew_Edge(edge))
    {
        if(--start_node->new_edge_count == 0)
            { *status |= SGL_US_FIN_START_NODE; }
    }

    if(start_node == end_node)
    {
        if(edge->add_loop_count++ == 0)
            { *status |= (SGL_US_NEW_EDGE | SGL_US_NEW_LOOP); }

        return edge;
    }

    edge->add_edge_count++;
    for(i = 0; i < sgl_getDegree_Edge(edge); i++)
    {
        SGL_Node *end_node_i = sgl_getEndNode_Edge(edge, i);
        
        if(end_node == end_node_i)
            { return edge; }
    }

    node_holder = (SGL_Node**)append_ArrayList(&edge->end_nodes);
   *node_holder = end_node;
    
   *status |= SGL_US_NEW_EDGE;
    return edge;
}

SGL_Edge* sgl_findNewEdge_Node(SGL_Node *node, int start_search_index)
{
    int i, j, k;
    SGL_Edge *edge;

    if(node->new_edge_count == 0)
        { return NULL; }

    if(node->new_edge_index == -1)
        { node->new_edge_index = 0; }
        
    i = start_search_index == -1 ? node->new_edge_index : start_search_index;

    for(j = 0; j < sgl_getDegree_Node(node); j++)
    {
        k = (i + j) % sgl_getDegree_Node(node);
        edge = sgl_getEdge_Node(node, k);

        if(sgl_isNew_Edge(edge))
        {
            node->new_edge_index = k;
            return edge;
        }
    }

    return NULL;
}

//------------------------------------------------------------------------------
// Graph Edge Functions
//------------------------------------------------------------------------------

SGL_Edge* sgl_alloc_Edge(void)
{
    SGL_Edge *edge = (SGL_Edge*)sgl_malloc(sizeof(SGL_Edge));
    assertion(edge != NULL, "sgl_alloc_Edge: out of memory");

    return edge;
}

void sgl_dealloc_Edge(SGL_Edge* edge)
{
    assertion(edge != NULL, "sgl_dealloc_Edge: NULL pointer");
    sgl_free(edge);
}

void sgl_init_Edge(SGL_Edge *edge, SGL_Node *start_node, SGL_StimulusId stimulus_id, SGL_Node *end_node)
{
    edge->start_node  = start_node;
    edge->stimulus_id = stimulus_id;

    edge->add_loop_count = 0;
    edge->add_edge_count = 0;

    edge->data = NULL;

    init_ArrayList(&edge->end_nodes, sizeof(SGL_Node*), SGL_END_NODES_ALLOC_BY, NULL);

    if(end_node == NULL)
        { return; }

    if(end_node != start_node)
    {
        SGL_Node **node_holder = (SGL_Node**)append_ArrayList(&edge->end_nodes);
       *node_holder = end_node;
    }
    else
    {
        edge->add_loop_count++;
    }
}

void sgl_initWithData_Edge(SGL_Edge *edge, SGL_Node *start_node, SGL_StimulusId stimulus_id,
    SGL_Node *end_node, void *data)
{
    sgl_init_Edge(edge, start_node, stimulus_id, end_node);
    edge->data = data;
}

void sgl_destroy_Edge(SGL_Edge *edge)
{
    SGL_Node *start_node = edge->start_node;
    SGL_DataDestructor destructor = start_node != NULL ? start_node->graph->edge_destructor : NULL;

    destroy_ArrayList(&edge->end_nodes);
    
    if(destructor != NULL)
        { destructor(edge->data); }
}

//------------------------------------------------------------------------------
// Graph Functions
//------------------------------------------------------------------------------

SGL_Graph* sgl_alloc_Graph(void)
{
    SGL_Graph *graph = (SGL_Graph*)sgl_malloc(sizeof(SGL_Graph));
    assertion(graph != NULL, "sgl_alloc_Graph: out of memory");

    return graph;
}

void sgl_dealloc_Graph(SGL_Graph* graph)
{
    assertion(graph != NULL, "sgl_dealloc_Graph: NULL pointer");
    sgl_free(graph);
}

void sgl_init_Graph
(
    SGL_Graph *graph,
    SGL_DataDestructor node_destructor,
    SGL_DataDestructor edge_destructor,
    SGL_DataDestructor graph_destructor
)
{
    sgl_init_Random(graph);
    sgl_setId_Graph(graph, 0);

    init_ArrayList(&graph->nodes, sizeof(SGL_Node), SGL_NODES_ALLOC_BY, (Destructor)sgl_destroy_Node);
    init_ArrayList(&graph->path, sizeof(SGL_Edge*), SGL_PF_PATH_ALLOC_BY, NULL);

    graph->hash_map = create_HashMap(&type_String, &type_Integer, SGL_NODES_HASH_SIZE, (HashFunc)hash_String);

    graph->node_destructor  = node_destructor;
    graph->edge_destructor  = edge_destructor;
    graph->graph_destructor = graph_destructor;

    graph->data = NULL;
}

void sgl_initWithData_Graph
(
    SGL_Graph *graph, void *data,
    SGL_DataDestructor node_destructor,
    SGL_DataDestructor edge_destructor,
    SGL_DataDestructor graph_destructor
)
{
    sgl_init_Graph(graph, node_destructor, edge_destructor, graph_destructor);
    graph->data = data;
}

void sgl_destroy_Graph(SGL_Graph *graph)
{
    SGL_DataDestructor destructor = graph->graph_destructor;

    destroy_ArrayList(&graph->nodes);
    destroy(graph->hash_map);

    if(destructor != NULL)
        { destructor(graph->data); }
}

SGL_Node* sgl_findNode_Graph(SGL_Graph *graph, SGL_NodeId *node_id)
{
    Integer *index;
    SGL_Node *node;

    if((index = get_HashMap(r(graph->hash_map), r(node_id->id))) == NULL)
        { return NULL; }

    node = (SGL_Node*)get_ArrayList(&graph->nodes, *index);
    destroy(index);

    return node;
}

SGL_Node* sgl_addNode_Graph(SGL_Graph *graph, SGL_NodeId *node_id, SGL_UpdateStatus *status)
{
    int id = graph->nodes.size;
    SGL_Node *node = sgl_findNode_Graph(graph, node_id);

   *status = SGL_US_OLD_ALL;

    if(node != NULL)
        { return node; }

    node = (SGL_Node*)append_ArrayList(&graph->nodes);
    sgl_init_Node(node, id, node_id, graph);

    put_HashMap(r(graph->hash_map), r(node_id->id), create_Integer(id));

    *status = SGL_US_NEW_NODE;
    return node;
}

SGL_Edge* sgl_addEdge_Graph(SGL_Graph *graph, SGL_NodeId *start_node_id,
    SGL_StimulusId stimulus_id, SGL_NodeId *end_node_id, SGL_UpdateStatus *status)
{
    SGL_Edge *edge;
    SGL_Node *start_node;
    SGL_Node *end_node;
    SGL_UpdateStatus add_status;

   *status = SGL_US_OLD_ALL;

    start_node = sgl_addNode_Graph(graph, start_node_id, &add_status);
   *status |= (add_status & SGL_US_NEW_START_NODE);

    end_node = end_node_id != NULL ? sgl_addNode_Graph(graph, end_node_id, &add_status) : NULL;
   *status |= (add_status & SGL_US_NEW_END_NODE);
    
    edge = sgl_addEdge_Node(start_node, stimulus_id, end_node, &add_status);
   *status |= add_status;
    
    return edge;
}

SGL_Node* sgl_findNonFinishedNode_Graph(SGL_Graph *graph)
{
    int i;

    for(i = 0; i < sgl_getSize_Graph(graph); i++)
    {
        SGL_Node *node = sgl_getNode_Graph(graph, i);
        
        if(!sgl_isFinished_Node(node))
            { return node; }
    }

    return NULL;
}

//------------------------------------------------------------------------------
// Path Finder Functions
//------------------------------------------------------------------------------

int sgl_targetNodeChecker(SGL_Node *current_node, SGL_Node *target_node)
{
    return current_node == target_node;
}

int sgl_nonFinishedNodeChecker(SGL_Node *current_node, SGL_Node *target_node)
{
    return current_node->new_edge_count;
}

int sgl_deterministicEdgeChecker(SGL_Edge *edge)
{
    return sgl_isValid_Edge(edge) && sgl_isDeterministic_Edge(edge);
}

int sgl_emptyEdgeChecker(SGL_Edge *edge)
{
    return sgl_isValid_Edge(edge);
}

static void sgl_init_PathFinder
(
    SGL_PathFinder *path_finder,
    SGL_Node       *source_node,
    SGL_Node       *target_node,
    int             hash_size
)
{
    SGL_Node **node_holder;
    
    path_finder->source_node  = source_node;
    path_finder->target_node  = target_node;
    
    // Initialize the list of nodes.
    init_ArrayList(&path_finder->nodes, sizeof(SGL_Node*), SGL_NODES_ALLOC_BY, NULL);

    // Append the source node into the list.
    node_holder = (SGL_Node**)append_ArrayList(&path_finder->nodes);
   *node_holder = source_node;

    // Initialize the hash map (source -> null).
    init_Int2PtrHashMap(&path_finder->nodes_map, hash_size, SGL_PF_MIN_HASH_SIZE);
    put_Int2PtrHashMap(&path_finder->nodes_map, source_node->id, SGL_PF_NULL_EDGE);

    // A node checker is used.
    if(target_node == NULL)
    {
        path_finder->nodes_queue = create_PQueue(SGL_NODES_ALLOC_BY, 0);

        source_node->distance = 0;
        insert_PQueue(path_finder->nodes_queue, source_node->weight, (void*)0);
    }
}

static void sgl_destroy_PathFinder(SGL_PathFinder *path_finder)
{
    destroy_ArrayList(&path_finder->nodes);
    destroy_Int2PtrHashMap(&path_finder->nodes_map);

    // A node checker is used.
    if(path_finder->target_node == NULL)
    {
        destroy_PQueue(path_finder->nodes_queue);
    }
}

static int sgl_getDistance_Edge(SGL_Edge *edge)
{
    const int loops = edge->add_loop_count;
    const int edges = edge->add_edge_count;

    // The average number of tries to exit the loop.
    return edges == 0 ? SGL_MAXIMAL_WEIGHT : 1 + (loops / edges);
}

static int sgl_getWeight_Edge(SGL_Edge *edge)
{
    int i;

    int distance   = sgl_getDistance_Edge(edge);
    int max_weight = SGL_MINIMAL_WEIGHT;

    for(i = 0; i < sgl_getDegree_Edge(edge); i++)
    {
        SGL_Node *end_node = sgl_getEndNode_Edge(edge, i);

        if(end_node->weight > max_weight)
            { max_weight = end_node->weight; }
    }

    // Weight is (max(end_node){weight(end_node)} + distance).
    return sgl_add_Weight(max_weight, distance);
}

void sgl_updateWeight_Node(SGL_Node *node, SGL_NodeChecker node_checker, SGL_EdgeChecker edge_checker)
{
    int i, j;
    int min_weight = SGL_MAXIMAL_WEIGHT;

    if(j = node_checker(/* source node*/ node, /* target node*/ NULL))
    {
        node->weight = SGL_INITIAL_WEIGHT;
        return;
    }

    for(i = 0; i < sgl_getDegree_Node(node); i++)
    {
        SGL_Edge *edge  = sgl_getEdge_Node(node, i);

        if(edge_checker(edge))
        {
            int edge_weight = sgl_getWeight_Edge(edge);

            if(edge_weight < min_weight)
                { min_weight = edge_weight; }
        }
    }

    // Weight min(edge){max(end_node){weight(end_node) + distance}}.
    node->weight = min_weight;
}

static void randomize_Order(SGL_Graph *graph, int *order, int size, int permutations)
{
    int i, j, k;

    // Generate the natural order.
    for(i = 0; i < size; i++)
        { order[i] = i; }

    if(size <= 1) { return; }

    // Apply a number of random permutations.
    for(i = 0; i < permutations; i++)
    {
        int temp = order[j = sgl_next_Random(graph) % size];
        order[j] = order[k = sgl_next_Random(graph) % size];
        order[k] = temp;
    }
}

static int sgl_findPath_Graph
(
    SGL_Graph      *graph,
    SGL_Node       *source_node,
    SGL_Node       *target_node,
    SGL_NodeChecker node_checker,
    SGL_EdgeChecker edge_checker,
    int             randomize
)
{
    int i, j, k, index;
    SGL_Node *node, **node_holder, *end_node;
    SGL_Edge *edge, **edge_holder;
    int verdict = 0;

    SGL_PathFinder path_finder;
    
    // Clear the path to be returned.
    clear_ArrayList(&graph->path);

    // The source node is a satisfactory one.
    if(node_checker(source_node, target_node))
        { return 1; }

    // Initialize the path finder.
    sgl_init_PathFinder(&path_finder, source_node, target_node,
        (graph->nodes.size < SGL_PF_MAX_HASH_SIZE ? graph->nodes.size : SGL_PF_MAX_HASH_SIZE));

    // In the beginning, the list of nodes contains only the source node.
    for(i = 0; i < path_finder.nodes.size; i++)
    {
        index = (target_node == NULL ? (int)extractMin_PQueue(path_finder.nodes_queue) : i);
        node = *(SGL_Node**)get_ArrayList(&path_finder.nodes, index);

        // Update the node weight.
        if(target_node == NULL)
            { sgl_updateWeight_Node(node, node_checker, edge_checker); }

        // The current node is satisfactory.
        if(node_checker(node, target_node))
        {
            edge = (SGL_Edge*)get_Int2PtrHashMap(&path_finder.nodes_map, node->id);

            // Construct the path in reverse order.
            for(;;)
            {
                // Append the edge into the path.
                edge_holder = (SGL_Edge**)append_ArrayList(&graph->path);
               *edge_holder = edge;

                // Update the node weight.
                if(target_node == NULL)
                    { sgl_updateWeight_Node(edge->start_node, node_checker, edge_checker); }

                edge = (SGL_Edge*)get_Int2PtrHashMap(&path_finder.nodes_map, edge->start_node->id);

                // The path has been successfully constructed.
                if(edge == SGL_PF_NULL_EDGE)
                    { break; }
            }

            verdict = 1;
            break;
        }
        // The current node is not satisfactory.
        else
        {
            // The order of traversal can be randomized.
            int size = sgl_getDegree_Node(node);
            int order[size];
            
            randomize_Order(graph, order, size, (randomize ? size / 2 : 0));
        
            // Expand the current node.
            for(j = 0; j < size; j++)
            {
                edge = sgl_getEdge_Node(node, order[j]);

                // Some edges (e.g., nondeterministic ones) can be skipped.
                if(!edge_checker(edge))
                    { continue; }

                for(k = 0; k < sgl_getDegree_Edge(edge); k++)
                {
                    end_node = sgl_getEndNode_Edge(edge, k);

                    // The pair (end_node -> edge) is put into the hash map (the node should be new).
                    if((SGL_Edge*)put_Int2PtrHashMap(&path_finder.nodes_map, end_node->id, edge) == edge)
                    {
                        if(target_node == NULL)
                        {
                            // Append the end node into the priority queue.
                            const int node_distance = node->distance;
                            const int edge_distance = sgl_getDistance_Edge(edge);
                            
                            end_node->distance = sgl_add_Weight(node_distance, edge_distance);

                            insert_PQueue(path_finder.nodes_queue,
                                sgl_add_Weight(end_node->weight, end_node->distance),
                                (void*)path_finder.nodes.size);
                        }

                        // Append the end node into the list.
                        node_holder = (SGL_Node**)append_ArrayList(&path_finder.nodes);
                       *node_holder = end_node;
                    }
                }
            }
        }
    }

    // Destroy the path finder.
    sgl_destroy_PathFinder(&path_finder);

    return verdict;
}

int sgl_findPath_TargetNode_Graph
(
    SGL_Graph      *graph,
    SGL_Node       *source_node,
    SGL_Node       *target_node,
    SGL_EdgeChecker edge_checker,
    int             randomize
)
{
    return sgl_findPath_Graph(graph, source_node, target_node, sgl_targetNodeChecker, edge_checker, randomize);
}

int sgl_findPath_NonFinishedNode_Graph
(
    SGL_Graph      *graph,
    SGL_Node       *source_node,
    SGL_EdgeChecker edge_checker,
    int             randomize
)
{
    return sgl_findPath_Graph(graph, source_node, NULL, sgl_nonFinishedNodeChecker, edge_checker, randomize);
}

int sgl_findPath_NodeChecker_Graph
(
    SGL_Graph      *graph,
    SGL_Node       *source_node,
    SGL_NodeChecker node_checker,
    SGL_EdgeChecker edge_checker,
    int             randomize
)
{
    return sgl_findPath_Graph(graph, source_node, NULL, node_checker, edge_checker, randomize);
}

//------------------------------------------------------------------------------
// Randomization Functions
//------------------------------------------------------------------------------

static unsigned reverse_Bits(unsigned i)
{
    i = (((i & 0xaaaaaaaa) >> 1) | ((i & 0x55555555) << 1));
    i = (((i & 0xcccccccc) >> 2) | ((i & 0x33333333) << 2));
    i = (((i & 0xf0f0f0f0) >> 4) | ((i & 0x0f0f0f0f) << 4));
    i = (((i & 0xff00ff00) >> 8) | ((i & 0x00ff00ff) << 8));

    return ((i >> 16) | (i << 16));
}

SGL_Edge* sgl_getNewEdge_Random(SGL_Graph *graph, SGL_Node *node)
{
    int i = node->new_edge_index == -1 ? reverse_Bits(graph->id) % sgl_getDegree_Node(node) : -1;

    return sgl_findNewEdge_Node(node, i);
}

SGL_Edge* sgl_getEdge_Random(SGL_Graph *graph, SGL_Node *node)
{
    int i = sgl_next_Random(graph) % sgl_getDegree_Node(node);
    return sgl_getEdge_Node(node, i);
}

//------------------------------------------------------------------------------
// Graph Printer Functions
//------------------------------------------------------------------------------

static void sgl_toFile_Edge(FILE *file, SGL_Graph *graph, SGL_Edge *edge, int full_graph)
{
    int i;

    SGL_Node *start_node = sgl_getStartNode_Edge(edge);
    int start_node_index = sgl_getId_Node(start_node);

    for(i = 0; i < sgl_getDegree_Edge(edge); i++)
    {
        SGL_Node *end_node = sgl_getEndNode_Edge(edge, i);
        int end_node_index = sgl_getId_Node(end_node);

        if(full_graph || end_node_index > start_node_index)
        {
        
#ifdef SGL_GRAPHVIZ_FORMAT
            fprintf(file, "  %d -> %d [label=%d];\n",
                start_node_index, end_node_index, sgl_getLabel_Edge(edge));
#else
            fprintf(file, "  <edge>\n");
            fprintf(file, "    <label>%d</label>\n",       sgl_getLabel_Edge(edge));
            fprintf(file, "    <end_node>%s</end_node>\n", sgl_getName_Node(end_node));
            fprintf(file, "  </edge>\n");
#endif
        }
    }
}

int sgl_toFile_Graph(SGL_Graph *graph, const char *filename, int argc, char **argv)
{
    int i, j;
    FILE *file;
    SGL_Edge *edge;
    SGL_Node *node;
    
    int full_graph = 1;

    for(i = 0; i < argc; i++)
    {
        if(!strcmp("--full-graph", argv[i]))
            { full_graph = 1; }
        else if(!strcmp("--forward-tree", argv[i]))
            { full_graph = 0; }
    }
    
    if((file = fopen(filename, "w")) == NULL)
        { return 0; }

#ifdef SGL_GRAPHVIZ_FORMAT
    fprintf(file, "digraph SGL {\n");

    if(sgl_getSize_Graph(graph) > 0)
    {
        node = sgl_getNode_Graph(graph, 0);

        fprintf(file, "  graph [root=%d];\n", node->id);
        fprintf(file, "  %d [label=\"%s\", shape=doublecircle];\n",
            sgl_getId_Node(node), sgl_getName_Node(node));
    }

    for(i = 1; i < sgl_getSize_Graph(graph); i++)
    {
        node = sgl_getNode_Graph(graph, i);

        fprintf(file, "  %d [label=\"%s\", shape=circle];\n",
            sgl_getId_Node(node), sgl_getName_Node(node));
    }
#else
    fprintf(file, "<graph>\n");
#endif

    for(i = 0; i < sgl_getSize_Graph(graph); i++)
    {
        node = sgl_getNode_Graph(graph, i);

        for(j = 0; j < sgl_getDegree_Node(node); j++)
        {
            edge = sgl_getEdge_Node(node, j);
            sgl_toFile_Edge(file, graph, edge, full_graph);
        }
    }

#ifdef SGL_GRAPHVIZ_FORMAT
    fprintf(file, "}\n");
#else
    fprintf(file, "</graph>\n");
#endif

    fclose(file);

    return 1;
}
